/********************************************************************************

Copyright (C) 2017 Lode Audio Ltd
support@lodeaudio.com
 
This file is part of Lode Audio - JavaScript Control Libraries.
 
This project can not be copied and/or distributed without the express
permission of Lode Audio Ltd
 
Author Ed Stark - Lode Audio

V1.5.0   -   Initial Beta Release 
V1.5.1   -   Production RC1
V1.5.2	 -	 Fixes to reloading browse from cache
V1.5.3	 -	 Fixes for missing browse items causing loop.
V1.5.4	 -	 Fixes for for empty browse results.

********************************************************************************/
function LODE_LOG(data){
		CF.log("LODE: " + data);
}
 
function LODE_LOG_OBJECT(data){
		//CF.logObject(data);
}

var LodeAudioServer = function(parameters)	{
		
	// Edit the settings here if you want to set the default values	
	var module = {
		defaultSysName: "Lode Audio Driver",
        sysName: "", //Passable parameter
		zoneName: "", //Passable parameter
		disableZoneGrouping: ""
    };
	
	/*************************************************************************************************
	
	Edit Joins Below
	
	*************************************************************************************************/
	
	module.currentVersion = "1.5.4";
	
	//Serial
	module.jCurrentTrackName = "s4001";
	module.jCurrentTrackArtist = "s4002";
	module.jCurrentTrackAlbum = "s4003";
	module.jCurrentTrackAlbumArt = "s4004";
	module.jTrackProgressStart = "s4005";
	module.jTrackProgressEnd = "s4006";
	module.jBackGroundBlurImage = "s4007"; //Seting this way out of the range as there is a global style issue when chaning it's opacity.
	module.jCurrentZoneName = "s4008";
	module.jCurrentBrowseTitle  = "s4009";
	
	
	module.jBrowsePlayOptionsAlbumArt = "s4010";
	module.jBrowsePlayOptionsTrackInfo = "s4011";
	module.jQueuePositionInfo = "s4012";
	module.jNextTrackInfo = "s4013";
	module.jSearchText = "s4014";
	module.jPlaylistName = "s4015";
	module.jQueueTitle = "s4016";
	
	//--- Buttons  ---//
	module.jNextTrack = "d4001";
	module.jPreviousTrack = "d4002";
	module.jPlay = "d4003";
	module.jMute = "d4004";
	module.jBrowseHome = "d4005";
	module.jBrowseBack = "d4006";	
	module.jShuffle = "d4007";
	module.jRepeat = "d4008";
	module.jClearQueue = "d4009";
	module.jSaveQueue = "d4010";
	module.jCloseBrowsePlayOptionsView= "d4011";
	module.jCloseZonesView = "d4012";
	module.jShowZonesView = "d4013";
	module.jPauseAll = "d4014";
	module.jToggleQueueAndBrowse = "d4015";
	module.jDoSearch = "d4016";
	
	module.jCancelQueueItemOptions = "d4017";
	module.jPlayQueueItemOption = "d4018";
	module.jUpQueueItemOption = "d4019";
	module.jDownQueueItemOption = "d4020";
	module.jDeleteQueueItemOption = "d4021";
	
	
	//-- View Control --//
	module.jViewTrackProgressView = "d4100";
	module.jBrowsePlayOptionsView = "d4101";
	module.jZonesView = "d4102";
	module.jShowBrowse = "d4103";
	module.jShowQueue = "d4104";
	module.jShowQueueItemOptions = "d4105";
	module.jShowSaveQueue = "d4106";
	module.jCancelSaveQueue = "d4107";
	module.jShowSearchView = "d4108";
	module.jCancelShowSearch = "d4109";
	module.jShowBrowseLoading = "d4110";
	
	module.connected = "d4200";
	module.disconnected = "d4201";
	
	//Analogue
	module.jVolumeVal = "a4001";
	module.jTrackProgress = "a4002";

	//Lists
	module.jZoneList = "l4001";
	module.jBrowseList = "l4002";
	module.jQueueList = "l4003";
	module.jBrowsePlayOPtionsList = "l4004";
	
	//Constants
	module.maxBrowseListItems = 1000;
	module.backgroundOpacity = 0.2;
	module.browseDisplayType = "LIST"; 
	module.browseGridColumnCount = 6;
	module.queueListStringLength = 50;
	module.browseListStringLength = 50;
	module.zonesPlayingStringLength = 25;
	module.groupedIntro = "Grouped with: ";
	
	//Sub page constants
	module.zoneListChildSubPage = "ZoneListItemChild";
	module.zoneListMasterCurrentSubPage = "ZoneListItemMasterCurrent";
	module.zoneListMasterSubPage = "ZoneListItemMaster";
	
	/*************************************************************************************************
	
	Module Start
	
	*************************************************************************************************/
	
	module.model = new Model(module);
	module.sysName = parameters["sysName"] || defaultSysName;
	module.zoneName = parameters["zoneName"] || "";
	module.disableZoneGrouping = parameters["disableZoneGrouping"] || false;
	
	module.currentBrowseObject = {};
	module.zoneLocking = false;
	
	module.blockVolumeUpdate = false;
	module.browseIsShowing = true;
	
	module.currentQueueItem = new QueueItem();
	module.currentBrowseItem = new BrowseItem();
	
	module.playOptionsData = [];
	module.browseListArray = [];
	module.browseInterupted = false;
	
	module.cancelQueueShortPress = false;
	module.cancelBrowseShortPress = false;
	
	module.zoneTrackUpdateEventRegistrations = [];
	
	module.currentQueueItemIndex = 0;
	module.currentSearchTerm = "";
	
	
	module.queueInitialised = false;
	module.searchCanceled = false;
	
	module.Startup = function() {
		
		LODE_LOG(module.defaultSysName + " " + module.currentVersion + " driver is loaded!");
		CF.watch(CF.FeedbackMatchedEvent, module.sysName, "IncomingData", module.incomingData);			
		CF.watch(CF.ConnectionStatusChangeEvent, module.sysName, module.onConnectionChange, true);	
		CF.watch(CF.GUIResumedEvent, module.sysName, module.onConnectionChange, true);
		
		//Internal event registrations
		if(!module.disableZoneGrouping) {
			module.model.registerZonesUpdateListener(module.updateZones);
		}
		
		module.model.getCurrentZone().registerTrackUpdateListener(function(){module.updateTrack()});
		module.model.registerBrowseUpdateListener(function(e){module.handleBrowse(e)});
		module.model.registerSearchUpdateListener(function(e){module.handleBrowse(e)});
		module.model.getCurrentZone().registerVolumeUpdateListener(module.setVolume);
		module.model.getCurrentZone().getTrack().registerTrackProgressListener(module.updateTrackProgress);
		module.model.getCurrentZone().registerNextTrackUpdateListener(function(){module.updateNextTrack()});
		module.model.getCurrentZone().registerTransportStateUpdateListener(module.updateTransportState);
		module.model.getCurrentZone().registerZoneUpdateListener(module.updateZone);
		module.model.getCurrentZone().registerMuteUpdateListener(module.updateMute);
		module.model.getCurrentZone().registerQueueUpdateListener(module.updateQueue);
		module.model.getCurrentZone().registerPlayModeUpdateListener(module.updatePlayMode);
		module.model.getCurrentZone().getQueue().registerCurrentQueueItemUpdateListener(module.updateCurrentQueueItem);
		
		//Watches for buttons presses
		CF.watch(CF.ObjectPressedEvent, module.jNextTrack, module.next);
		CF.watch(CF.ObjectPressedEvent, module.jPreviousTrack, module.previous);
		CF.watch(CF.ObjectPressedEvent, module.jPlay, module.play);
		CF.watch(CF.ObjectPressedEvent, module.jMute, module.mute);
		CF.watch(CF.ObjectPressedEvent, module.jBrowseHome, module.browseHome);
		CF.watch(CF.ObjectPressedEvent, module.jBrowseBack, module.browseBack);
		CF.watch(CF.ObjectPressedEvent, module.jShuffle, module.shuffle);
		CF.watch(CF.ObjectPressedEvent, module.jRepeat, module.repeat);
		CF.watch(CF.ObjectPressedEvent, module.jClearQueue, module.clearQueue);
		CF.watch(CF.ObjectPressedEvent, module.jShowSaveQueue, module.showSaveQueue);
		CF.watch(CF.ObjectPressedEvent, module.jSaveQueue, module.saveQueue);
		CF.watch(CF.ObjectPressedEvent, module.jCloseBrowsePlayOptionsView, function(e){CF.setJoin(module.jBrowsePlayOptionsView,0)});
		CF.watch(CF.ObjectPressedEvent, module.jCloseZonesView, function(e){CF.setJoin(module.jZonesView,0)});
		CF.watch(CF.ObjectPressedEvent, module.jShowZonesView, function(e){CF.setJoin(module.jZonesView,1)});
		CF.watch(CF.ObjectPressedEvent, module.jPauseAll, module.pauseAll);
		CF.watch(CF.ObjectPressedEvent, module.jShowSearchView, function(e){CF.setJoin(module.jShowSearchView,1)});
		CF.watch(CF.ObjectPressedEvent, module.jCancelShowSearch, function(e){CF.setJoin(module.jShowSearchView,0)});
		CF.watch(CF.ObjectPressedEvent, module.jToggleQueueAndBrowse, module.toggleQueueAndBrowse);
		
		CF.watch(CF.ObjectPressedEvent, module.jCancelQueueItemOptions, module.cancelQueueItemOptions);
		CF.watch(CF.ObjectPressedEvent, module.jCancelSaveQueue, module.cancelSaveQueue);
		CF.watch(CF.ObjectPressedEvent, module.jPlayQueueItemOption, module.playQueueItemOption);
		CF.watch(CF.ObjectPressedEvent, module.jUpQueueItemOption, module.upQueueItemOption);
		CF.watch(CF.ObjectPressedEvent, module.jDownQueueItemOption, module.downQueueItemOption);
		CF.watch(CF.ObjectPressedEvent, module.jDeleteQueueItemOption, module.deleteQueueItemOption);

		CF.watch(CF.ObjectPressedEvent, module.jVolumeVal, function(e){module.blockVolumeUpdate = true;});
		CF.watch(CF.ObjectDraggedEvent, module.jVolumeVal, module.updateVolume);
		CF.watch(CF.ObjectReleasedEvent, module.jVolumeVal, function(e){module.blockVolumeUpdate = false;});
		CF.watch(CF.ObjectDraggedEvent, module.jTrackProgress, module.seek);		
		
		CF.watch(CF.OrientationChangeEvent, module.onOrientationChange, true);
		
		CF.setProperties({join:module.jBackGroundBlurImage, opacity:module.backgroundOpacity});
		CF.setJoin(module.jShowBrowse,1);
	};
	
	module.incomingData = function (itemName, matchedString) {
		module.model.handleData(matchedString);
	};
				
	module.onConnectionChange = function (system, connected, remote) {
		if (connected) {
			module.model.serverConnected();
			module.model.init(100);
			
			if(module.zoneName !== ""){
				module.zoneLocking = true;
				LODE_LOG("Zone name specified, setting to: " + module.zoneName);
				module.model.setCurrentZone(module.zoneName);
			}
					
		} else {
			LODE_LOG("Disconnected.")									
		}
	};	
	
	module.onOrientationChange = function(pageName, newOrientation) {
		LODE_LOG("Orientation of page " + pageName + " changed to: " + newOrientation);
		if (newOrientation == CF.LandscapeOrientation) {
		} else {
		}
	 };
	
	module.pauseAll = function(){
		for (var zone in module.model.getZones()) {
            module.model.getZone(zone).sendPause();
        }
	};
	
	module.doSearch = function(){
		LODE_LOG("In do search");
		//LODE_LOG_OBJECT(module.model.getSearchableSources());
		module.searchCanceled = false;
		CF.getJoin(module.jSearchText, function(join, value) {
			if (value != null) {
				module.browseInterupted = true;
				for (var index in module.model.getSearchableSources()) {
					var source = module.model.getSearchableSources()[index];
					//LODE_LOG_OBJECT(source);
					if(source.path.indexOf("sire") != -1 && value != ""){
						//LODE_LOG("Search Criteria:" + source.criteria);
						for(var i = 0; i < source.criteria.length; i++){
							module.model.getCurrentZone().sendSearch(source.path,source.criteria[i][0],value,0,5);
						}
					}
				}
			}
			
		});
		
		CF.setJoin(module.jShowSearchView, 0);
		CF.setJoin(module.jShowBrowseLoading, 1);
		if(!module.browseIsShowing){
			module.toggleQueueAndBrowse();
		}
	};
	
	module.toggleQueueAndBrowse = function(){
		if(module.browseIsShowing){
			CF.setJoin(module.jToggleQueueAndBrowse, 1);
			CF.setJoin(module.jShowBrowse, 0);
			CF.setJoin(module.jShowQueue, 1);
			module.browseIsShowing = false;
		}
		else{
			CF.setJoin(module.jToggleQueueAndBrowse, 0);
			CF.setJoin(module.jShowBrowse, 1);
			CF.setJoin(module.jShowQueue, 0);
			module.browseIsShowing = true;
		}
		
	};
	
	module.hideQueueItemOptions = function(){
		CF.setJoin(module.jShowQueueItemOptions, 0);
	};
	
	module.cancelQueueItemOptions = function(){
		module.hideQueueItemOptions();
	};
	
	module.playQueueItemOption = function(){
		module.model.getCurrentZone().sendPlayQueueItem(module.currentQueueItemIndex);
		module.hideQueueItemOptions();
	};
	
	module.upQueueItemOption = function(){
		module.model.getCurrentZone().sendReorderTrackInQueue(module.currentQueueItemIndex, module.currentQueueItemIndex-1);
		module.hideQueueItemOptions();
	};
	
	module.downQueueItemOption = function(){
		module.model.getCurrentZone().sendReorderTrackInQueue(module.currentQueueItemIndex, module.currentQueueItemIndex+1);
		module.hideQueueItemOptions();
	};
	
	module.deleteQueueItemOption = function(){
		module.model.getCurrentZone().sendRemoveFromQueue(module.currentQueueItemIndex);
		module.hideQueueItemOptions();
	};
	
	module.playOptionsListItemSelected = function(listIndex) {
		module.playOptionsData[listIndex][0](module.currentBrowseItem.getRes());
		CF.setJoin(module.jBrowsePlayOptionsView, 0);
	};
	
	module.queueListItemLongPress = function(index){
		LODE_LOG("Queue Item Long Press");
		module.currentQueueItem = module.model.getCurrentZone().getQueue().getQueueItems()[index];
		module.currentQueueItemIndex = index;
		module.cancelQueueShortPress = true;
		CF.setJoin(module.jShowQueueItemOptions, 1);
	};
	
	module.queueListItemShortPress = function(index){
		if(!module.cancelQueueShortPress) {
			LODE_LOG("Queue Item Short Press")
			module.model.getCurrentZone().sendPlayQueueItem(index);
			module.currentQueueItem = module.model.getCurrentZone().getQueue().getQueueItems()[index];
			module.currentQueueItemIndex = index;
			LODE_LOG('Selected queueItem: ' + module.currentQueueItem.getTitle());
		}
		module.cancelQueueShortPress = false;
	};
	
	
	module.browseListItemLongPress = function(index, position, showPlayOptions){
		var item = index;
		module.currentBrowseItem = module.currentBrowseObject.getBrowseItemForIndex(item);
		LODE_LOG('Selected Long Press browseItem: ' + module.currentBrowseItem.getTitle());
		module.cancelBrowseShortPress = true;
		if(module.currentBrowseObject.getBrowseItemForIndex(item).isPlayable()) {
			module.updateBrowsePlayOptions();
			CF.setJoin(module.jBrowsePlayOptionsView, 1);
		}
	};
	
	module.browseListItemShortPress = function(index, position, showPlayOptions){
		if(!module.cancelBrowseShortPress) {
			var item = index;
			LODE_LOG("Browse items selected at index: " + item);
			module.currentBrowseItem = module.currentBrowseObject.getBrowseItemForIndex(item);
			LODE_LOG('Selected Short Press browseItem: ' + module.currentBrowseItem.getTitle());
			
			if(module.currentBrowseObject.getBrowseItemForIndex(item).isContainer()) {
				module.searchCanceled = true;
				module.currentBrowseObject.getBrowseItemForIndex(item).browseInto();
				CF.listRemove(module.jBrowseList);
				CF.setJoin(module.jShowBrowseLoading, 1);
			}
			else{
				module.updateBrowsePlayOptions();
				CF.setJoin(module.jBrowsePlayOptionsView, 1);
			}
		}
		module.cancelBrowseShortPress = false;
	};
	
	
	module.browseListItemPressAndHold = function(index, position, showPlayOptions){
		
		LODE_LOG("Browse item press and hold");
		var item = index;
		
		module.currentBrowseItem = module.currentBrowseObject.getBrowseItemForIndex(item);
		LODE_LOG('Selected browseItem: ' + module.currentBrowseItem.getTitle());
		
		if(module.currentBrowseObject.getBrowseItemForIndex(index).isPlayable()){
			module.updateBrowsePlayOptions();
			CF.setJoin(module.jBrowsePlayOptionsView, 1);
		}
		
	};
	
	
	module.updateBrowsePlayOptions = function(){
		CF.setJoin(module.jBrowsePlayOptionsAlbumArt, module.currentBrowseItem.getAlbumArt());
		CF.setJoin(module.jBrowsePlayOptionsTrackInfo, module.currentBrowseItem.getTitle() + '\r\n' + module.currentBrowseItem.getArtist());
		
		module.playOptionsData = [];
		
		if(module.currentBrowseItem.isPlayable()){
			module.playOptionsData.push([module.model.getCurrentZone().sendPlayNow, "Play Now"]);
		}

		if(this.currentBrowseItem.isQueueable()){
			module.playOptionsData.push([module.model.getCurrentZone().sendPlayNext, "Play Next"]);
			module.playOptionsData.push([module.model.getCurrentZone().sendReplaceQueue, "Replace Queue"]);
			module.playOptionsData.push([module.model.getCurrentZone().sendAddToQueue, "Add To Queue"]);
		}

		if(this.currentBrowseItem.isFavouriteable()){
			module.playOptionsData.push([module.model.getCurrentZone().sendAddToFav, "Add to Favourites"]);
		}
		
		if(this.currentBrowseItem.isFavourite()){
			module.playOptionsData.push([module.model.getCurrentZone().sendDeleteFav, "Delete Favourite"]);
		}

		if(this.currentBrowseItem.isPlaylist()){
			module.playOptionsData.push([module.model.getCurrentZone().sendDeletePlaylist, "Delete Playlist"]);
		}
		
		var listArray = [];
		for(var i = 0; i < module.playOptionsData.length; i++){
			
			listArray.push({
				s4001: module.playOptionsData[i][1]
			});
		}
		
		CF.listRemove(module.jBrowsePlayOPtionsList);
		CF.listAdd(module.jBrowsePlayOPtionsList, listArray);
		
	};
	
	module.partyMode = function(){
		module.model.getCurrentZone().sendPartyMode();
	};
	
	module.addMember = function(listJoin){
		LODE_LOG('Zone item selected: ' + listJoin);
		CF.getJoin(listJoin, function(j,v,t) {
			module.model.getCurrentZone().sendAddMember(t["[zoneName]"]);
		});
	};
	
	module.removeMember = function(listJoin){
		LODE_LOG('Zone item selected: ' + listJoin);
		CF.getJoin(listJoin, function(j,v,t) {
			module.model.getCurrentZone().sendRemoveMember(t["[zoneName]"]);
		});
	};
	
	module.zoneListItemSelected = function(listJoin){
		// listJoin example: l2:0:d1
		LODE_LOG('Zone item selected: ' + listJoin);
		
		CF.getJoin(listJoin, function(j,v,t) {
			LODE_LOG('Changing zone to selected: ' + t["[zoneName]"]);
			module.model.setCurrentZone(t["[zoneName]"]);
		});
		
	};
	
	module.clearQueue = function(){
		module.model.getCurrentZone().sendClearQueue();
	};
	
	module.showSaveQueue = function() {
		CF.setJoin(module.jShowSaveQueue, 1);
	};
	
	module.saveQueue = function() {
		CF.getJoin(module.jPlaylistName, function(join, value) {
			if (value != null) {
				module.model.getCurrentZone().sendSaveQueue(value);
				CF.setJoin(module.jPlaylistName,"");
			}
			
		});
		module.cancelSaveQueue();
	};
	
	module.cancelSaveQueue = function() {
		CF.setJoin(module.jShowSaveQueue, 0);
	};
	
	module.next = function(){
		module.model.getCurrentZone().sendNextTrack();
	};
	
	module.previous = function(){
		module.model.getCurrentZone().sendPreviousTrack();
	};
	
	module.mute = function(){
		module.model.getCurrentZone().sendMute(module.model.getCurrentZone().getMute() ? 0 : 1);
	};
	
	module.browseHome = function(){
		if(!module.currentBrowseObject.isRoot()){
			CF.listRemove(module.jBrowseList);
			CF.setJoin(module.jShowBrowseLoading, 1);
			module.searchCanceled = true;
			module.model.browseHome();	
		}
	};
	
	module.browseBack = function(){
		if(!module.currentBrowseObject.isRoot()){
			module.browseInterupted = true;
			module.searchCanceled = true;
			CF.listRemove(module.jBrowseList);
			CF.setJoin(module.jShowBrowseLoading, 1);
			module.model.browseBack();
		}
	};
	
	module.shuffle = function(){
		
		CF.getJoins([module.jRepeat, module.jShuffle], function(joins) {

			var mode = 'NORMAL';
			var isContinuous = 0;
			var isShuffle = 0;
		
			if (joins[module.jRepeat].value != null) {
				isContinuous = joins[module.jRepeat].value;
			}
			
			if (joins[module.jShuffle].value != null) {
				isShuffle = joins[module.jShuffle].value;
			}
			
			LODE_LOG('Repeat presses, isContinuous:' +  isContinuous + ' isShuffle:' + isShuffle);
		
			if (isContinuous == 0 && isShuffle == 0) {
				mode = 'SHUFFLE_NOREPEAT';
			}
			else if (isContinuous == 1 && isShuffle == 1) {
				mode = 'REPEAT_ALL';
			}
			else if (isShuffle == 0 && isContinuous == 1) {
				mode = 'SHUFFLE';
			}
		  
			module.model.getCurrentZone().sendPlayMode(mode);
		});

	};
	
	module.repeat = function(){
		
		CF.getJoins([module.jRepeat, module.jShuffle], function(joins) {

			var mode = 'NORMAL';
			var isContinuous = 0;
			var isShuffle = 0;
		
			if (joins[module.jRepeat].value != null) {
				isContinuous = joins[module.jRepeat].value;
			}
			
			if (joins[module.jShuffle].value != null) {
				isShuffle = joins[module.jShuffle].value;
			}
			
			LODE_LOG('Repeat presses, isContinuous:' +  isContinuous + ' isShuffle:' + isShuffle);
		
			if (isContinuous == 0 && isShuffle == 0) {
				mode = 'REPEAT_ALL';
			}
			else if (isContinuous == 1 && isShuffle == 1) {
				mode = 'SHUFFLE_NOREPEAT';
			}
			else if (isShuffle == 1 && isContinuous == 0) {
				mode = 'SHUFFLE';
			}
		  
			module.model.getCurrentZone().sendPlayMode(mode);
		});
		
	};
	
	module.play = function(){
		if(module.model.getCurrentZone().getTransportState() == "PLAYING"){
            module.model.getCurrentZone().sendPause();
        }
        else{
            module.model.getCurrentZone().sendPlay();
        }
	};
	
	module.updateVolume = function(join, value, tokens, tags){
		LODE_LOG("Volume from slider: " + value);
		var newVal = Math.floor((value / 65535) * 100);
		module.model.getCurrentZone().sendVolume(newVal);
	};
	
	module.seek = function(join, value, token, tags)
	{
		var newVal = Math.floor((value / 65535) * 100);
		module.model.getCurrentZone().sendSeek(newVal);
	}
	
	
	module.updatePlayMode = function(){
		LODE_LOG('In update playmode function.');
		var mode = module.model.getCurrentZone().getPlayMode();
       
		if (mode == 'REPEAT_ALL') {
			CF.setJoin(module.jRepeat, 1);
            CF.setJoin(module.jShuffle, 0);
        }
        if (mode == 'NORMAL') {
			CF.setJoin(module.jRepeat, 0);
            CF.setJoin(module.jShuffle, 0);
        }
        if (mode == 'SHUFFLE') {
			CF.setJoin(module.jRepeat, 1);
            CF.setJoin(module.jShuffle, 1);
        }
        if (mode == 'SHUFFLE_NOREPEAT') {
			CF.setJoin(module.jRepeat, 0);
            CF.setJoin(module.jShuffle, 1);
        }	
		
	};
	
	module.updateCurrentQueueItem = function(){
		var currentId = module.model.getCurrentZone().getQueue().getCurrentQueueItem();
		
		if (currentId > 0) {
			currentId = currentId - 1;
		}
			
		if(module.model.getCurrentZone().getQueue().size() > currentId &&  module.queueInitialised) {
			
			CF.listScroll(module.jQueueList, currentId, CF.TopPosition, true);
			
			CF.listUpdate(module.jQueueList, [
				{ index:CF.AllItems, d4001:0 }
			]);
			
			CF.listUpdate(module.jQueueList, [
				{ index:currentId, d4001:1 }
			]);
			
			
			CF.setJoin(module.jQueuePositionInfo, (currentId+1) + " of " +  module.model.getCurrentZone().getQueue().size());
			if(module.model.getCurrentZone().getTrack().isAStream())
			{
				CF.setJoin(module.jQueuePositionInfo, "");
			}
		}
		
	};
	
	module.updateMute = function(){
		if(module.model.getCurrentZone().isFixedVolume()) {
			//this.muteBtn.style.visibility = "hidden";
		}
		else {
			var state = module.model.getCurrentZone().getMute();
			CF.setJoin(module.jMute, state);
		}
	};
	
	module.updateZone = function() {
		CF.setJoin(module.jCurrentZoneName, module.model.getCurrentZone().getName());
	};
	
	module.updateTrackProgress = function(){
		CF.setJoin(module.jViewTrackProgressView, 1);
		CF.setJoin(module.jTrackProgressStart, module.model.getCurrentZone().getTrack().getDuration());
		CF.setJoin(module.jTrackProgressEnd, module.model.getCurrentZone().getTrack().getRemaining());
		CF.setJoin(module.jTrackProgress,(module.model.getCurrentZone().getTrack().getProgress()/100)*65535);
	};
	
	module.setVolume = function(){
		if(!module.blockVolumeUpdate){
			var newVal = Math.floor((module.model.getCurrentZone().getVolume()/100)*65535); 
			LODE_LOG("Sending analogue volume to slider: " + newVal);
			CF.setJoin(module.jVolumeVal, newVal);
		}
	};
	
	module.updateNextTrack = function()
	{
		LODE_LOG("In next track callback, got: " + module.model.getCurrentZone().getNextTrack());
		if(module.model.getCurrentZone().getNextTrack() != undefined && module.model.getCurrentZone().getNextTrack() != "")
		{
			CF.setJoin(module.jNextTrackInfo, "Next: " + module.model.getCurrentZone().getNextTrack());
		}
		else 
		{
			module.clearNextTrack();
		}
	};
	
	
	module.clearNextTrack = function()
	{
		CF.setJoin(module.jNextTrackInfo, "");	
	}
	
	module.updateTrack = function(){
		CF.setJoin(module.jCurrentTrackName, module.model.getCurrentZone().getTrack().getName());	
		CF.setJoin(module.jCurrentTrackArtist, module.model.getCurrentZone().getTrack().getArtist());	
		CF.setJoin(module.jCurrentTrackAlbum, module.model.getCurrentZone().getTrack().getAlbum());	
		CF.setJoin(module.jCurrentTrackAlbumArt, module.model.getCurrentZone().getTrack().getAlbumArt());	
		CF.setJoin(module.jBackGroundBlurImage, module.model.getCurrentZone().getTrack().getAlbumArt() + "?blur=true");
		LODE_LOG("Is track a stream: " + module.model.getCurrentZone().getTrack().isAStream());
		CF.setJoin(module.jViewTrackProgressView, !module.model.getCurrentZone().getTrack().isAStream());
		if(module.model.getCurrentZone().getTrack().isAStream())
		{
			CF.setJoin(module.jQueuePositionInfo, "");
			module.clearNextTrack();
		}

	};
	
	module.updateTransportState = function(){
		switch(module.model.getCurrentZone().getTransportState()) {
            case 'PLAYING':
                CF.setJoin(module.jPlay,1);
                break;
            case 'STOPPED':
            case 'PAUSED_PLAYBACK':
                CF.setJoin(module.jPlay,0);
                break;
        };
	};
	
	var zoneViewUpdateEvent = function(index, zone){
		this.zoneObj = module.model.getZone(zone);
		this.index = index;
		
		var obj = this;
		
		this.update = function(){
			LODE_LOG('Updateing track for zone: ' + obj.zoneObj.getName() + ' track name: ' + obj.zoneObj.getTrack().getName());
			CF.listUpdate(module.jZoneList, [
				{ 
					index:obj.index, 
					s4002:module.truncateString(obj.zoneObj.getTrack().getName(), module.zonesPlayingStringLength), 
					s4003:module.truncateString(obj.zoneObj.getTrack().getArtist(), module.zonesPlayingStringLength), 
					s4004: obj.zoneObj.getTrack().getAlbumArt()
				}
			]);
		}
		
		this.deRegister = function(){
			obj.zoneObj.deregisterTrackUpdateListener(obj.update);
		}
		
		obj.zoneObj.registerTrackUpdateListener(obj.update);
		
	}
	
	module.updateZones = function(){
		var listArray = [];
		var i = 0;
		for(var zone in module.model.getZones()){
			
			var zoneObj = module.model.getZone(zone);
			var title = zoneObj.getName();
			var groupedString = (zoneObj.getParent() ? module.groupedIntro + " "  + zoneObj.getParent().getName() : "" );
			var subpage = module.zoneListMasterSubPage;
			
			if(!zoneObj.isMaster() && zoneObj.getParent().getName() == module.model.getCurrentZone().getName()){
				subpage = module.zoneListChildSubPage;
			}
			
			if(zoneObj.isMaster() && module.model.getCurrentZone().getName() == title){	
				subpage = module.zoneListMasterCurrentSubPage;
			}
			
			if(!zoneObj.isMaster()){
				zoneObj = zoneObj.getParent();	
			}
			
			listArray.push({
				s4001: title,
				s4002: module.truncateString(zoneObj.getTrack().getName(), module.zonesPlayingStringLength),
				s4003: module.truncateString(zoneObj.getTrack().getArtist(), module.zonesPlayingStringLength),
				s4004: zoneObj.getTrack().getAlbumArt(),
				s4005: groupedString,
				d4001: {
					tokens: {
						"[zoneName]": zoneObj.getName()
					}
				},
				d4002: {
					tokens: {
						"[zoneName]": title
					}
				},
				subpage: subpage,
				}
			);
			
			if(module.zoneTrackUpdateEventRegistrations[title]){
				module.zoneTrackUpdateEventRegistrations[title].deRegister();
			}
			
			module.zoneTrackUpdateEventRegistrations[title] = new zoneViewUpdateEvent(i, zoneObj.getName());
			
					
			i++;
		}	
		
		CF.listRemove(module.jZoneList);
		CF.listAdd(module.jZoneList, listArray);
		
	};
	

	
	module.truncateString = function(str, length, ending) {
		if (length == null) {
		  length = 100;
		}
		if (ending == null) {
		  ending = '...';
		}
		if (str.length > length) {
		  return str.substring(0, length - ending.length) + ending;
		} else {
		  return str;
		}
	};
	
	module.updateQueue = function(){
		var queue = module.model.getCurrentZone().getQueue();
		
		LODE_LOG('Queue updated item count: ' + queue.size());
		if(queue.size() > 0) {
			CF.setJoin(module.jQueueTitle, "Queue has " + queue.size() + " item(s)");
		}
		else{
			CF.setJoin(module.jQueueTitle, "Queue");
		}
		
		var listArray = [];
		
		for(var i = 0; i < queue.size(); i++){
		
			var item = queue.getQueueItems()[i];
			listArray.push({
				s4001: module.truncateString(item.getTitle(), module.queueListStringLength), 
				s4002: module.truncateString(item.getArtist(), module.queueListStringLength),
				s4003: item.getAlbumArt() + "?size=200"
			});
		}
		
		CF.listRemove(module.jQueueList);
		CF.listAdd(module.jQueueList, listArray);
		module.queueInitialised = true;
		module.model.getCurrentZone().queryCurrentQueueItem();
		
	};
	
	module.handleSearch = function(browseObject){
		
		var params = browseObject.getParentId().split("/");
		LODE_LOG("In search for term: " + params[4]);
		CF.getJoin(module.jSearchText, function(join, value) {
			if (value != null && value == params[4]) {
				
				if(module.currentSearchTerm != value)
				{
					CF.listRemove(module.jBrowseList);
					module.currentBrowseObject = new Browse();
				}
				
				var browseListArray = [];
				
				for(var i = 0; i < browseObject.size(); i++){
			
					var browseItem = browseObject.getBrowseItemForIndex(i);
					
					browseListArray.push({
						s4001: module.truncateString(browseItem.getTitle(),module.browseListStringLength),
						s4002: module.truncateString(browseItem.getArtist(),module.browseListStringLength),
						s4003: browseItem.getAlbumArt(),
						d4001: browseItem.isPlayable() && browseItem.isContainer() ? 1 : 0
					});
					
					module.currentBrowseObject.addBrowseItem(browseItem);
					
					LODE_LOG("Adding to array: " + browseItem.getTitle());
				}
				
				CF.listAdd(module.jBrowseList, browseListArray);
				module.currentSearchTerm = value;
			}
		});
		
		CF.setJoin(module.jShowBrowseLoading, 0);
	};
	
	module.handleBrowse = function(browseObject){
		
		LODE_LOG("View controller received new browse");
		CF.setJoin(module.jShowBrowseLoading, 0);
		
		if(browseObject.getParentId().indexOf("searchresult") != -1 && !module.searchCanceled){
			module.handleSearch(browseObject);
			return;
		}
		
		//Reset the current search term as we are no longer in search.
		//A bit dirty, will be removed once browse stack is handled outside of the model.
		module.currentSearchTerm = "";
		
		if(!browseObject.isRoot() && browseObject.getParentId() != module.currentBrowseItem.getId() && !module.browseInterupted)
		{
			LODE_LOG("Browse was unexpected, exiting view update.")
			return;
		}
		
		
		module.browseInterupted = false;
		
		var title = "";
		if(module.model.getBrowseStack().length > 0){
			title = module.model.getBrowseStack()[module.model.getBrowseStack().length - 1].getTitle();
		}
		else{
			title = "Music";
		}
		
		CF.setJoin(module.jCurrentBrowseTitle, title);
		
		var browseListArray = [];
		
		if(browseObject.isRoot() || module.currentBrowseObject.getParentId() != browseObject.getParentId())
		{	
			LODE_LOG("Creating new list array");
		}
		
		LODE_LOG("Browse object size: "  + browseObject.size());
		
		for(var i = browseObject.getLastRequestedIndex(); i < browseObject.size(); i++){
			
			var browseItem = browseObject.getBrowseItemForIndex(i);
			
			browseListArray.push({
				index: i,
				title: (browseItem.getTitle() != "" && browseItem.getArtist() == ""),
				s4001: module.truncateString(browseItem.getTitle(),module.browseListStringLength),
				s4002: module.truncateString(browseItem.getArtist(),module.browseListStringLength),
				s4003: browseItem.getAlbumArt(),
				d4001: browseItem.isPlayable() && browseItem.isContainer() ? 1 : 0
			});
			
			LODE_LOG("Adding to array: " + browseItem.getTitle());
		}

		
		if(browseObject.isRoot() || module.currentBrowseObject.getParentId() != browseObject.getParentId())
		{
			LODE_LOG("New browse, removing and adding.");
			CF.listRemove(module.jBrowseList);
			CF.listAdd(module.jBrowseList, browseListArray);
			module.currentBrowseObject = browseObject;
		}
		else
		{
			LODE_LOG("Update to browse, just updating list.");
			CF.listUpdate(module.jBrowseList, browseListArray);
		}
		
		if(browseObject.getLoadedCount() < browseObject.getTotalMatches() && browseObject.getLoadedCount() <= (module.maxBrowseListItems - module.model.browseRequestAmount)){
			browseObject.requestNextItems();
		}
		
		LODE_LOG("Finished processing browse");
		
	};
	
	return module;
};




var Model = function(module){

    var minApiBuild = "1.5.10";

	var zones = {};
    var serverConnectedListeners = [];
    var serverDisconnectedListeners = [];
    var zonesUpdateListeners = [];
    var browseUpdateListeners = [];
    var searchUpdateListeners = [];
    var currentBrowseTitleUpdateListeners = [];
    var browseRequestWaitListeners = [];
    var searchableSources = {};
    var browseCache = {};
    var searchCache = {};
    var currentZone = "";
    var currentZoneObject = new CurrentZone();
    var browseStack = [];
	var module = module;
	var browseRequestAmount = 0;
	this.currentBrowseRequestId
	
	var api = new API(module);
    var obj = this;

	
    this.serverConnected = function(){
        LODE_LOG("Model has connected to Lode unit.");
        api.sendInitialisationQueries();
    };
   
	this.getApi = function(){
		return api;
	};

	this.sendCommand = function(command){
		api.send(command);
	};

	this.handleData = function(data){
		api.handleData(data);
	};
	
    this.init = function(browseRequestAmount){

		LODE_LOG("Calling Model init()");

		api.registerZoneEventListener(function(e){obj.setZones(e)});
		api.registerVersionEventListener(function(e){obj.versionCheck(e)});
		api.registerTransportStateEventListener(function(e){obj.getZone(e[0]).setTransportState(e[1])});
		api.registerTrackEventListener(function(e){obj.getZone(e[0]).updateTrack(e[0], "", e[3], e[1], e[2], e[4], e[7]);});
		api.registerNextTrackEventListener(function(e){obj.getZone(e[0]).setNextTrack(e[1]);});
		api.registerTrackProgressEventListener(function(e){obj.getZone(e[0]).getTrack().updateProgress(e[3], e[4], e[6]);});
		api.registerVolumeEventListener(function(e){obj.getZone(e[0]).setVolume(e[1])});
		api.registerMuteEventListener(function(e){obj.getZone(e[0]).setMute(parseInt(e[1]))});
		api.registerCurrentQueueItemEventListener(function(e){obj.getZone(e[0]).getQueue().setCurrentQueueItem(e[1])});
		api.registerQueueChangedEventListener(function(e){api.queryQueueForZone(e[0])});
		api.registerSearchCritereaEventListener(function(e){obj.setSearchableSources(e)});
		api.registerPlayModeEventListener(function(e){obj.getZone(e[0]).setPlayMode(e[1])});
		api.registerMessageEventListener(function(e){obj.getCurrentZone().fireServerErrorMessageEvents(e)});
		api.registerBrowseEventListener(function(e){obj.handleBrowse(e)});
		api.registerQueueEventListener(function(e){obj.handleQueue(e)});
		this.browseRequestAmount = browseRequestAmount;
        api.init(browseRequestAmount);
    };

    this.versionCheck = function(data){
        if(this.versionCompare(data[1].replace("-SNAPSHOT",""),minApiBuild) >= 0) {
            obj.fireServerConnectedEvents();
        }
        else{
            obj.fireServerDisconnectedEvents("Please update your Lode unit firmware in order to use the desktop controller software.");
        }
    };

    this.getCurrentZone = function(){
        return currentZoneObject;
    };

	this.addZone = function(zone){
        if(!zones[zone.getName()]){
            zones[zone.getName()] = zone;
        }
    };

    this.getZone = function(zone){
        return zones[zone];
    };

    this.hasZone = function(zone){
        var obj = zones[zone];
        return !!obj;
    };

    this.clearZoneRelationships = function(){
        for(var zone in zones){
            this.getZone(zone).setMaster(false);
            this.getZone(zone).children = [];
        }
    };
	
    this.setCurrentZone = function(zone){
        if (currentZone != zone) {
            LODE_LOG("Changing zone context to: " + zone);
            currentZone  =  zone;
			var thisZone =  this.getZone(currentZone);
			if(thisZone){
				currentZoneObject.update(thisZone);
			}
			else{
				this.addZone(new Zone(zone,api, this));
				currentZoneObject.update(this.getZone(zone));
				this.browseHome();
			}
            this.fireZonesUpdateEvents();
        }
    };

    this.setZones = function(data){

        var newZones = {};

        var allZonesList = [];
        for(var i = 0; i < data.length; i++){
            for(var j = 0; j < data[i].length; j++){
                allZonesList.push(data[i][j]);
            }
        }

        for(var zone in zones){
            var i = allZonesList.indexOf(zone);
            if(i != -1){
                newZones[zone] = this.getZone(zone);
            }
        }

        zones = newZones;

        for(var i = 0; i < data.length; i++){

            var zone = this.getZone(data[i][0]);

            if(!zone) {
                zone = new Zone(data[i][0], api, this);
                this.addZone(zone);
                
            }
            zone.update(data[i]);
			this.queryZone(zone.getName());
        }

        if (currentZone == "" || allZonesList.indexOf(currentZone) == -1) {
            this.setCurrentZone(data[0][0]);
            this.browseHome();
        }

        this.fireZonesUpdateEvents();
    };

    this.getZones = function(){
        return zones;
    };

    this.handleQueue = function(data){

        var queue = this.getZone(data[0]).getQueue();
        if(queue) {
            queue.setQueueItems([]);

            if(parseInt(data[1]) > 0) {
                for (var i = 2; i < data.length; i++) {
                    var queueItem = new QueueItem(data[i][0], data[i][1], data[i][2], data[i][3], data[i][4]);
                    queue.addQueueItem(queueItem);
                }
            }

        }

        this.getZone(data[0]).fireQueueUpdateEvents();
    };

    this.queryZone = function (zone) {
		api.queryQueueForZone(zone);
        api.queryTrackForZone(zone);
		api.queryNextTrackForZone(zone);
        api.queryPlayModeForZone(zone);
        api.queryVolumeForZone(zone);
        api.queryTransportStateForZone(zone);
        api.queryMuteForZone(zone);
        api.queryCurrentQueueItemForZone(zone);
    };

    this.queryChildVolume = function(zone){
        api.queryMuteForZone(zone);
        api.queryVolumeForZone(zone);
    };



    this.setSearchableSources = function(data){
        //LODE_LOG("Searchable Sources:" + data);
        searchableSources = {};
        for (var i = 0; i < data.length - 2; i = i + 3) {
            var searchSource = new SearchableSource(data[i]);
			//LODE_LOG_OBJECT(searchSource);
            searchSource.setCriteria(data[i + 2]);
            this.addSearchAbleSource(searchSource);
        }
    };

    this.addSearchAbleSource = function(source){
        if(!searchableSources[source.path]){
            searchableSources[source.path] = source;
        }
    };

    this.getSearchableSources = function(){
        return searchableSources;
    };


    this.getBrowseObject = function(id){
        var browseObject = browseCache[id];
        if(browseObject){
            return browseObject;
        }
        else{
            return "";
        }
    };

    this.getSearchObject = function(id){
        var searchObject = searchCache[id];
        if(searchObject){
            return searchObject;
        }
        else{
            return "";
        }
    };

    this.clearSearchCache = function(){
        searchCache = {};
    };

    this.deleteFromBrowseCache = function(item){
        delete browseCache[item.getRes()];
    };

    this.clearBrowseCache = function(){
        browseCache = [];
    };

    this.addToBrowseStack = function(browse){
        browseStack.push(browse);
        try {
            LODE_LOG("Adding item to browse stack." + browse.getParent().getParentId());
        }
        catch(e){
            LODE_LOG("ERROR Adding to browse stack, no parent.");
        }
        this.fireCurrentBrowseTitleUpdateEvents();
    };

    this.getBrowseStack = function(){
        return browseStack;
    };

    this.clearBrowseStack = function(){
        browseStack = [];
    };

    this.browseHome = function(){
		LODE_LOG('---- Browsing Home ------');
        this.clearBrowseCache();
        this.clearBrowseStack();
        this.sendBrowse("sire:///", 0);
    };

    this.browseBack = function(){
		LODE_LOG('---- Browsing Back ------');
		LODE_LOG('BrowseStack Length: ' + browseStack.length);
		LODE_LOG(browseStack);
        if(browseStack.length > 1) {
            var last = browseStack[browseStack.length - 1];
            this.deleteFromBrowseCache(last);
            this.loadBrowseStackItem(last.getParent());
        }
        else {
            this.browseHome();
        }
    };

    this.loadBrowseStackItem = function(object){
		object.setLastRequestedIndex(0);
        for(var i = (browseStack.length-1); i >= 0; i--){
            if(browseStack[i].getRes() != object.getParentId()){
                browseStack.pop();
            }
            else{
                break;
            }
        }

        if(object.getParentId().indexOf("searchresult") != -1) {
            this.fireSearchUpdateEvents(object);
        }
        else{
            this.fireBrowseUpdateEvents(object);
        }

    };

    this.handleSearch = function(data) {

        var searchObject = {};

        if (data.length > 3 && data[3][0] != "") {

            var sid = data[0];
            searchObject = searchCache[sid];

            if (!searchObject) {
                this.clearSearchCache();
                searchObject = new Browse(sid, data[1], data[2], this);
                searchCache[searchObject.getParentId()] = searchObject
            }

            searchObject.setIsSearchResult(true);

            for (var i = 3; i < data.length; i++) {
                var browseItem = new BrowseItem(data[i][0],data[i][1],data[i][2] != 'null' ? data[i][2] : '',data[i][3], data[i][4], data[i][5], searchObject, this);
                searchObject.addBrowseItem(browseItem);
            }


            var serviceId = api.getServiceIdForBrowseId(data[3][5]);

            if (serviceId) {

                var params = searchObject.getParentId().split("/");

                var title = searchObject.getTotalMatches() + ' ' + params[3];

                if (searchObject.getNumberReturned() > 1 && params[3].substr(params[3].length - 1).toLowerCase() != 's') {
                    title = title + 's';
                }

                var res = serviceId + "dosearch/" + params[3] + '/' + encodeURI(params[4]) + '/';
                var title = api.services[serviceId].getTitle() + " found " + title.toLowerCase().replace(/(^|\s)([a-z])/g, function (m, p1, p2) {
                        return p1 + p2.toUpperCase();
                    });
                var artist = "See all results.....";
                var albumArt = api.services[serviceId].getAlbumArt();
                var attributes = ['CONTAINER'];

                var browseItem = new BrowseItem(title, title, artist, albumArt, attributes, res, searchObject, this);
                browseItem.setIsaTitle(true);

                searchObject.addBrowseItemToStart(browseItem);
                searchObject.setNumberReturned(searchObject.getNumberReturned() + 1);

            }
        }

        this.fireSearchUpdateEvents(searchObject);

    };

    this.handleBrowse = function(data) {

        try {

            LODE_LOG("Model handling browse for:" + data[0] + " got " + data[1]);
			var havenew = false;
			
		    if(data[0].indexOf('sire://searchresults/') != -1){
                this.handleSearch(data);
                return;
            }

			if(this.currentBrowseRequestId != data[0] || this.currentBrowseRequestId == ""){
                return;
            }
			
            var browseObject = browseCache[data[0]];

            if (!browseObject || browseObject.isExpectingAnUpdate()) {
				LODE_LOG("Model received new browse, creating new object.");
                browseObject = new Browse(data[0], data[1], data[2], this);
                browseCache[data[0]] = browseObject
            }
			else{
				LODE_LOG("Model found browse in cache, appending new items.");
			}

            if(browseObject.size() < browseObject.getTotalMatches()) {
                for (var i = 0; i < browseObject.getTotalMatches(); i++) {
					LODE_LOG('adding browsItem:' + i);
                    var browseItem = new BrowseItem();
                    browseObject.addBrowseItem(browseItem);
                }
            }

            for (var i = 3; i < data.length; i++) {
                if(browseObject.getLoadedCount() < browseObject.size()) {
					//LODE_LOG('adding browse item: ' + data[i][1]);
					LODE_LOG("Model processing new browse item:, adding to browseObject at position: " +  browseObject.getLoadedCount());
                    var browseItem = browseObject.getBrowseItemForIndex(browseObject.getLoadedCount());
					browseItem.update(data[i][0], data[i][1], data[i][2] != 'null' ? data[i][2] : '', data[i][3], data[i][4], data[i][5], browseObject, this);
                    browseObject.incrementLoadedCount();
					havenew = true;
                }
            }

            if (browseObject.getParentId() == "sire:///") {
                api.setServices(browseObject);
                browseObject.setIsRoot(true);
            }

        } catch (e) {
            LODE_LOG('Error building browse object:' + e.message);
        }

		//if(havenew){
			this.fireBrowseUpdateEvents(browseObject);
		//}

    };


    this.versionCompare = function(v1, v2, options) {
        var lexicographical = options && options.lexicographical,
            zeroExtend = options && options.zeroExtend,
            v1parts = v1.split('.'),
            v2parts = v2.split('.');

        function isValidPart(x) {
            return (lexicographical ? /^\d+[A-Za-z]*$/ : /^\d+$/).test(x);
        }

        if (!v1parts.every(isValidPart) || !v2parts.every(isValidPart)) {
            return NaN;
        }

        if (zeroExtend) {
            while (v1parts.length < v2parts.length) v1parts.push("0");
            while (v2parts.length < v1parts.length) v2parts.push("0");
        }

        if (!lexicographical) {
            v1parts = v1parts.map(Number);
            v2parts = v2parts.map(Number);
        }

        for (var i = 0; i < v1parts.length; ++i) {
            if (v2parts.length == i) {
                return 1;
            }

            if (v1parts[i] == v2parts[i]) {
                continue;
            }
            else if (v1parts[i] > v2parts[i]) {
                return 1;
            }
            else {
                return -1;
            }
        }

        if (v1parts.length != v2parts.length) {
            return -1;
        }

        return 0;
    };

    ///// API Calls  ////

    this.sendBrowse = function(id, index){
        this.fireBrowseRequestWaitEvents(id);
		this.currentBrowseRequestId = id;
        api.queryBrowse(currentZone, id, index);
    };

    //////// EVENTS /////////

    this.registerZonesUpdateListener = function(listener){
        zonesUpdateListeners.push(listener);
    };

    this.deregisterZonesUpdateListener = function(listener){
        zonesUpdateListeners.splice(zonesUpdateListeners.indexOf(listener), 1);
    };

    this.fireZonesUpdateEvents = function(){
        for(var e = 0; e < zonesUpdateListeners.length; e++){
            try{
                zonesUpdateListeners[e]();
            }
            catch(err){
                LODE_LOG("Callback for zones update failed for " +  err.message);
            }
        }
    };

    this.registerBrowseUpdateListener = function(listener){
        browseUpdateListeners.push(listener);
    };

    this.deregisterBrowseUpdateListener = function(listener){
        browseUpdateListeners.splice(browseUpdateListeners.indexOf(listener), 1);
    };

    this.fireBrowseUpdateEvents = function(browseId){
        for(var e = 0; e < browseUpdateListeners.length; e++){
            try{
                browseUpdateListeners[e](browseId);
            }
            catch(err){
                LODE_LOG("Callback for browse failed for " + err.message);
            }
        }
    };


    this.registerCurrentBrowseTitleUpdateListener = function(listener){
        currentBrowseTitleUpdateListeners.push(listener);
    };

    this.deregisterCurrentBrowseTitleUpdateListener = function(listener){
        currentBrowseTitleUpdateListeners.splice(currentBrowseTitleUpdateListeners.indexOf(listener), 1);
    };

    this.fireCurrentBrowseTitleUpdateEvents = function(browseId){
        for(var e = 0; e < currentBrowseTitleUpdateListeners.length; e++){
            try{
                currentBrowseTitleUpdateListeners[e](browseId);
            }
            catch(err){
                LODE_LOG("Callback for browse failed for " + err.message);
            }
        }
    };

    this.registerSearchUpdateListener = function(listener){
        searchUpdateListeners.push(listener);
    };

    this.deregisterSearchUpdateListener = function(listener){
        searchUpdateListeners.splice(searchUpdateListeners.indexOf(listener), 1);
    };

    this.fireSearchUpdateEvents = function(browseId){
        for(var e = 0; e < searchUpdateListeners.length; e++){
            try{
                searchUpdateListeners[e](browseId);
            }
            catch(err){
                LODE_LOG("Callback for search failed for " + err.message);
            }
        }
    };


    this.registerServerConnectedListener = function(listener){
        serverConnectedListeners.push(listener);
    };

    this.deregisterServerConnectedListener = function(listener){
        serverConnectedListeners.splice(serverConnectedListeners.indexOf(listener), 1);
    };

    this.fireServerConnectedEvents = function(browseId){
        for(var e = 0; e < serverConnectedListeners.length; e++){
            try{
                serverConnectedListeners[e](browseId);
            }
            catch(err){
                LODE_LOG("Callback for server connection failed for " + err.message);
            }
        }
    };

    this.registerServerDisconnectedListener = function(listener){
        serverDisconnectedListeners.push(listener);
    };

    this.deregisterServerDisconnectedListener = function(listener){
        serverDisconnectedListeners.splice(serverDisconnectedListeners.indexOf(listener), 1);
    };

    this.fireServerDisconnectedEvents = function(browseId){
        for(var e = 0; e < serverDisconnectedListeners.length; e++){
            try{
                serverDisconnectedListeners[e](browseId);
            }
            catch(err){
                LODE_LOG("Callback for server disconnection failed for " + err.message);
            }
        }
    };

    this.registerBrowseRequestWaitListener = function(listener){
        browseRequestWaitListeners.push(listener);
    };

    this.deregisterBrowseRequestWaitListener = function(listener){
        browseRequestWaitListeners.splice(browseRequestWaitListeners.indexOf(listener), 1);
    };

    this.fireBrowseRequestWaitEvents = function(id){
        for(var e = 0; e < browseRequestWaitListeners.length; e++){
            try{
                browseRequestWaitListeners[e](id);
            }
            catch(err){
                LODE_LOG("Callback for browse request listener failed for " + err.message);
            }
        }
    };


};

var SearchableSource = function(path){

    this.path = path;
    this.criteria = [];

    this.setCriteria = function(someCriteria){
        LODE_LOG("Setting criteria for:" + this.path + " criteria: " + someCriteria);
        for(var i = 0; i < someCriteria.length;i=i+2){
                this.criteria.push([someCriteria[i], someCriteria[i+1]]);
                LODE_LOG("Adding criteria for:" + this.path + " criteria: " + someCriteria[i] + "    "  + someCriteria[i+1]);
        }

        LODE_LOG(this.criteria );
    };

    this.size = function(){
        return this.criteria.length;
    };


};

var Browse = function(parentId, totalMatches, numberReturned, model){

	var Model = model;
    var parentId = parentId;
    var totalMatches = parseInt(totalMatches);
    var numberReturned = parseInt(numberReturned);
    var isSearchResult = false;
    var browseItems = [];
    var isaRoot = false;
    var loadedCount = 0;
    var expectingAnUpdate = false;
    var lastRequestedIndex = 0;
    var lastScrollPosition = 0;

    this.getParentId = function(){
        return parentId;
    };

    this.setParentId = function(aparentId){
        parentId = aparentId;
    };

    this.getTotalMatches = function(){
        return totalMatches;
    };

    this.getNumberReturned = function(){
        return numberReturned;
    };

    this.setNumberReturned = function(num){
        numberReturned = num;
    };

	this.getLastRequestedIndex = function(){
		return lastRequestedIndex;
	};
	
	
	this.setLastRequestedIndex = function(num){
		lastRequestedIndex = num;
	};

    this.size = function(){
        return browseItems.length;
    };

    this.setBrowseItems = function(items){
        browseItems = items;
    };

    this.addBrowseItem= function(item){
        browseItems.push(item);
    };

    this.addBrowseItemToStart = function(item){
        browseItems.unshift(item);
    };

    this.removeBrowseItemAtIndex = function(index){
        browseItems.splice(index,1);
    };

    this.getBrowseItemForIndex = function(index){
        return browseItems[index];
    };

    this.getBrowseItems =  function(){
        return browseItems;
    };

    this.setIsSearchResult = function(isit){
        isSearchResult = isit;
    };

    this.isaSearchResult = function(){
        return isSearchResult;
    };

    this.isRoot = function(){
        return isaRoot;
    };

    this.setIsRoot = function(isit){
        isaRoot = isit;
    };

    this.setLoadedCount = function(count){
        loadedCount = count;
    };

    this.getLoadedCount = function(){
        return loadedCount;
    };

    this.getLastScrollPosition = function(){
        return lastScrollPosition;
    };

    this.setLastScrollPosition = function(slastScrollPosition){
        lastScrollPosition = slastScrollPosition;
    };

    this.incrementLoadedCount = function(){
        loadedCount++
    };

    this.isExpectingAnUpdate = function(){
        return expectingAnUpdate;
    };

    this.expectUpdate = function(){
        LODE_LOG("Browse object " + parentId + " is awaiting an update");
        expectingAnUpdate = true;
    };

    this.requestNextItems = function(){
        if(loadedCount < totalMatches) {
            Model.sendBrowse(parentId, loadedCount);
			lastRequestedIndex = loadedCount - 1;
        }
    };

     this.requestItemsFromIndex = function(index){
        LODE_LOG('In browse requestItemsFromIndex index=' + index + '  lastRequestedIndex=' + lastRequestedIndex);
        if(index % numberReturned == 0){
            var remaining = index - lastRequestedIndex;
            if(remaining > 0){
                Model.sendBrowse(parentId, lastRequestedIndex+numberReturned);
            }
            lastRequestedIndex = index;
        }

    };

};




var BrowseItem = function(id, title, artist, albumArt, attributes, res, parent, model) {
	
	var Model = model;
    var id = id || "";
    var title = title || "";
    var artist = artist || "";
    var albumArt = albumArt || "";
    var attributes = attributes || [];
    var res = res || "";
    var parent = parent;
    var isTitle = false;

    var browseItemUpdateListeners = [];

    this.update = function(aid, atitle, aartist, aalbumArt, aattributes, ares, aparent, model){
        id = aid;
        title = atitle;
        artist = aartist;
        albumArt = aalbumArt;
        attributes = aattributes;
        res = ares;
        parent = aparent;
        this.fireBrowseItemUpdateEvents();
		Model = model;
    };

    this.isLoaded = function(){
        return (id != "" && title != "");
    };

    this.getId = function(){
        return id;
    };

    this.getTitle = function(){
        return title;
    };

    this.getArtist = function(){
        return artist;
    };

    this.getAlbumArt = function(){
        return albumArt;
    };

    this.getAttributes = function(){
        return attributes;
    };

    this.getRes = function(){
        return res;
    };

    this.isaTitle = function(){
        return isTitle;
    };

    this.setIsaTitle = function(isit){
        isTitle = isit;
    };

    this.getParent = function(){
        return parent;
    };

    this.isPlayable = function(){
        return attributes.indexOf('PLAYABLE') != -1;
    };

    this.isPlaylist = function(){
        return attributes.indexOf('PLAYLIST') != -1;
    };

    this.isQueueable = function(){
        return attributes.indexOf('QUEUEABLE') != -1;
    };

    this.isFavouriteable = function(){
        return attributes.indexOf('FAVOURITEABLE') != -1;
    };

    this.isFavourite = function(){
        return attributes.indexOf('FAVOURITE') != -1;
    };

	this.isContainer = function(){
        return attributes.indexOf('CONTAINER') != -1;
    };


    this.registerBrowseItemUpdateListener = function(listener){
        browseItemUpdateListeners.push(listener);
    };

    this.deregisterBrowseItemUpdateListener = function(listener){
        browseItemUpdateListeners.splice(browseItemUpdateListeners.indexOf(listener), 1);
    };

    this.browseInto = function(){
		LODE_LOG('in browseInto');
        Model.addToBrowseStack(this);
        Model.sendBrowse(res, 0);
    };

    this.fireBrowseItemUpdateEvents = function(){
        for(var e = 0; e < browseItemUpdateListeners.length; e++){
            try{
                browseItemUpdateListeners[e](this);
            }
            catch(err){
                LODE_LOG("Callback for browse item failed for " + err.message);
            }
        }
    };

};




var CurrentQueue = function() {

    var queue = new Queue();
    var currentQueueItemUpdateListeners = [];

    this.update = function(currentQueue){

        if(queue != currentQueue){

            queue.deregisterCurrentQueueItemUpdateListener(this.fireCurrentQueueItemEvents);

            queue = currentQueue;

            queue.registerCurrentQueueItemUpdateListener(this.fireCurrentQueueItemEvents);

        }

    };

    this.getCurrentQueueItem = function(){
        return queue.getCurrentQueueItem();
    };

    this.size = function(){
        return queue.size();
    };

    this.getQueueItems = function(){
        return queue.getQueueItems();
    };

    this.registerCurrentQueueItemUpdateListener = function(listener){
        currentQueueItemUpdateListeners.push(listener);
    };

    this.fireCurrentQueueItemEvents = function(){
        for(var i = 0; i < currentQueueItemUpdateListeners.length; i++){
            try{
                currentQueueItemUpdateListeners[i]();
            }
            catch(err){
                LODE_LOG("Callback for current queue item failed for " +  err.message);
            }
        }
    };
};



var CurrentTrack = function(){

    var trackProgressListeners = [];
    var track = new Track();

    this.update = function(currentTrack){

        if(track != currentTrack){

            track.deregisterTrackProgressListener(this.fireTrackProgressEvents);

            track = currentTrack;

            track.registerTrackProgressListener(this.fireTrackProgressEvents);

            this.fireTrackProgressEvents();
        }
    };

    this.getZone = function(){
        return track.getZone();
    };

    this.getId = function(){
        return track.getId();
    };

    this.getName = function(){
        return track.getName();
    };

    this.getAlbum = function(){
        return track.getAlbum();
    };

    this.getArtist = function(){
        return track.getArtist();
    };

    this.getAlbumArt = function(){
        return track.getAlbumArt();
    };

	this.isAStream = function(){
		return track.isAStream();
	};
	
    this.getProgress = function(){
        return track.getProgress();
    };

    this.getDuration = function(){
        return track.getDuration();
    };

    this.getRemaining = function(){
        return track.getRemaining();
    };

    this.registerTrackProgressListener = function(listener){
        trackProgressListeners.push(listener);
    };

    this.fireTrackProgressEvents = function(){
        for(var e = 0; e <  trackProgressListeners.length; e++){
            try{
                trackProgressListeners[e]();
            }
            catch(e){
                LODE_LOG("Callback for track progress failed for " +  e.message);
            }
        }
    };
};



var CurrentZone = function(){

    var zone = new Zone();
    var volumeUpdateListeners = [];
    var muteUpdateListeners = [];
    var transportStateUpdateListeners = [];
    var playModeUpdateListeners = [];
    var queueUpdateListeners = [];
    var trackUpdateListeners = [];
	var nextTrackUpdateListeners = [];
    var zoneUpdateListeners = [];
    var serverErrorMessageListeners = [];
    var track = new CurrentTrack();
    var queue = new CurrentQueue();

    this.update = function(currentZone){

        if(zone != currentZone){

			LODE_LOG("Current Zone Object:");
			LODE_LOG_OBJECT(currentZone);
			
            //Deregister events
            zone.deregisterMuteUpdateListener(this.fireMuteUpdateEvents);
            zone.deregisterPlayModeUpdateListener(this.firePlayModeUpdateEvents);
            zone.deregisterQueueUpdateListener(this.fireQueueUpdateEvents);
            zone.deregisterTrackUpdateListener(this.fireTrackUpdateEvents);
			zone.deregisterNextTrackUpdateListener(this.fireNextTrackUpdateEvents);
            zone.deregisterTransportStateUpdateListener(this.fireTransportStateUpdateEvents);
            zone.deregisterVolumeUpdateListener(this.fireVolumeUpdateEvents);
            zone.deregisterZoneUpdateListener(this.fireZoneUpdateEvents);

            zone = currentZone;

            //Add new child proxies
            track.update(zone.getTrack());
            queue.update(zone.getQueue());

            //Register for events
            zone.registerMuteUpdateListener(this.fireMuteUpdateEvents);
            zone.registerPlayModeUpdateListener(this.firePlayModeUpdateEvents);
            zone.registerQueueUpdateListener(this.fireQueueUpdateEvents);
            zone.registerTrackUpdateListener(this.fireTrackUpdateEvents);
			zone.registerNextTrackUpdateListener(this.fireNextTrackUpdateEvents);
            zone.registerTransportStateUpdateListener(this.fireTransportStateUpdateEvents);
            zone.registerVolumeUpdateListener(this.fireVolumeUpdateEvents);
            zone.registerZoneUpdateListener(this.fireZoneUpdateEvents);

            //Fire events as zone has changed
            this.fireMuteUpdateEvents();
            this.firePlayModeUpdateEvents();
            this.fireQueueUpdateEvents();
            this.fireTrackUpdateEvents();
			this.fireNextTrackUpdateEvents();
            this.fireTransportStateUpdateEvents();
            this.fireVolumeUpdateEvents();
            this.fireZoneUpdateEvents();

            //Fire this here rather than in the children so we know the parent events have fired first.
            //Can't set the current queueItem before the queue has been updated.
            queue.fireCurrentQueueItemEvents();
        }
    };


    //Used for debug
    this.id = zone.getName();

    this.getName = function(){
        return zone.getName();
    };

    this.getVolume = function(){
        return zone.getVolume();
    };

    this.isFixedVolume = function(){
        return zone.isFixedVolume();
    };

    this.getMute = function(){
        return zone.getMute();
    };


    this.getTransportState = function(){
        return zone.getTransportState();
    };


    this.getPlayMode = function(){
        return  zone.getPlayMode();
    };

    this.hasChild = function(zone){

        for(var i = 0; i < this.getChildren().length; i++){
            if(this.getChildren()[i].getName() == zone){
                return true;
            }
        }
        return false;
    };

    this.getChildren = function(){
        return zone.getChildren();
    };

	this.getParent = function(){
		return zone.getParent();
	};

    this.getChildrenAndParent = function(){
        var zonesArr = [];
        zonesArr.push(this);

        for(var zone in this.getChildren()){
            zonesArr.push(this.getChildren()[zone]);
        }

        return zonesArr;
    };

    this.getQueue = function(){
        return queue;
    };

    this.getTrack = function(){
        return track;
    };

    this.getNextTrack = function(){
        return zone.getNextTrack();
    };

    this.isMaster = function(){
        return zone.isMaster();
    };

    ///////////   API Commands   /////////

    this.sendPlay = function(){
        zone.sendPlay()
    };

    this.sendPause = function(){
        zone.sendPause();
    };

    this.sendPreviousTrack = function(){
        zone.sendPreviousTrack();
    };

    this.sendNextTrack = function(){
        zone.sendNextTrack();
    };

    this.sendSeek = function(percentage){
        zone.sendSeek(percentage);
    };

    this.sendPlayNow = function(res){
        zone.sendPlayNow(res);
    };

    this.sendPlayNext = function(res){
        zone.sendPlayNext(res);
    };

    this.sendReplaceQueue = function(res){
        zone.sendReplaceQueue(res);
    };

    this.sendAddToQueue = function(res){
        zone.sendAddToQueue(res);
    };

    this.sendAddToFav = function(res){
        zone.sendAddToFav(res);
    };

    this.sendDeleteFav = function(res){
        zone.sendDeleteFav(res);
    };

    this.sendDeletePlaylist = function(res){
        zone.sendDeletePlaylist(res);
    };

    this.sendMute = function(state){
        zone.sendMute(state);
    };

    this.sendVolume = function(newVolume, throttled){
        zone.sendVolume(newVolume, throttled);
    };

    this.sendGroupedVolume = function(newVolume){
        zone.sendGroupedVolume(newVolume);
    };

    this.sendQueryVolume = function(){
        zone.sendQueryVolume();
    };

    this.sendPartyMode = function(){
        zone.sendPartyMode();
    };

    this.sendPlayMode = function(mode){
        zone.sendPlayMode(mode);
    };

    this.sendRemoveMember = function(member){
        zone.sendRemoveMember(member);
    };

    this.sendAddMember = function(member){
        zone.sendAddMember(member);
    };

    this.sendClearQueue = function(){
        zone.sendClearQueue();
    };

    this.removeAllMembers = function(){
        zone.removeAllMembers();
    };

    this.sendSaveQueue = function(playlistName){
        zone.sendSaveQueue(playlistName);
    };

	this.queryCurrentQueueItem = function(){
		zone.queryCurrentQueueItem();
	};
	
    this.sendReorderTrackInQueue = function(item, newPos){
        zone.sendReorderTrackInQueue(item, newPos);
    };

    this.sendPlayQueueItem = function(item){
        zone.sendPlayQueueItem(item);
    };

    this.sendRemoveFromQueue = function(item){
        zone.sendRemoveFromQueue(item);
    };

    this.sendSearch = function(path, criteria, term, index, total){
        zone.sendSearch(path, criteria, term, index, total);
    };


    //////////   Events   //////////
    this.registerTransportStateUpdateListener = function(listener){
        transportStateUpdateListeners.push(listener);
    };

    this.fireTransportStateUpdateEvents = function(){
        for(var e = 0; e < transportStateUpdateListeners.length; e++){
            try{
                transportStateUpdateListeners[e]();
            }
            catch(err){
                LODE_LOG("Callback for transport failed for " +  err.message);
            }
        }
    };

    this.registerVolumeUpdateListener = function(listener){
        volumeUpdateListeners.push(listener);
    };

    this.fireVolumeUpdateEvents = function(){
        for(var e = 0; e < volumeUpdateListeners.length; e++){
            try{
                volumeUpdateListeners[e]();
            }
            catch(err){
                LODE_LOG("Callback for volume failed for " + err.message);
            }
        }
    };

    this.registerMuteUpdateListener = function(listener){
        muteUpdateListeners.push(listener);
    };

    this.fireMuteUpdateEvents = function(){
        for(var e = 0; e < muteUpdateListeners.length; e++){
            try{
                muteUpdateListeners[e]();
            }
            catch(err){
                LODE_LOG("Callback for mute failed for " +  err.message);
            }
        }
    };

    this.registerPlayModeUpdateListener = function(listener){
        playModeUpdateListeners.push(listener);
    };

    this.firePlayModeUpdateEvents = function(){
        for(var e = 0; e < playModeUpdateListeners.length; e++){
            try{
                playModeUpdateListeners[e]();
            }
            catch(err){
                LODE_LOG("Callback for play mode failed for " +  err.message);
            }
        }
    };

    this.registerQueueUpdateListener = function(listener){
        queueUpdateListeners.push(listener);
    };

    this.fireQueueUpdateEvents = function(){
        for(var e = 0; e < queueUpdateListeners.length; e++){
            try{
                queueUpdateListeners[e]();
            }
            catch(err){
                LODE_LOG("Callback for queue failed for " +  err.message);
            }
        }
    };

    this.registerTrackUpdateListener = function(listener){
        trackUpdateListeners.push(listener);
    };

    this.fireTrackUpdateEvents = function(){
        for(var e = 0; e < trackUpdateListeners.length; e++){
            try{
                trackUpdateListeners[e]();
            }
            catch(err){
                LODE_LOG("Callback for track update failed for " +  err.message);
            }
        }
    };

	 this.registerNextTrackUpdateListener = function(listener){
        nextTrackUpdateListeners.push(listener);
    };

    this.fireNextTrackUpdateEvents = function(){
        for(var e = 0; e < nextTrackUpdateListeners.length; e++){
            try{
                nextTrackUpdateListeners[e]();
            }
            catch(err){
                LODE_LOG("Callback for next  track update failed for " +  err.message);
            }
        }
    };


    this.registerZoneUpdateListener = function(listener){
        zoneUpdateListeners.push(listener);
    };

    this.fireZoneUpdateEvents = function(){
        for(var e = 0; e < zoneUpdateListeners.length; e++){
            try{
                zoneUpdateListeners[e]();
            }
            catch(err){
                LODE_LOG("Callback for track update failed for " +  err.message);
            }
        }
    };


    this.registerServerErrorMessageListener = function(listener){
        serverErrorMessageListeners.push(listener);
    };

    this.deregisterServerErrorMessageListener = function(listener){
        serverErrorMessageListeners.splice(serverErrorMessageListeners.indexOf(listener), 1);
    };

    this.fireServerErrorMessageEvents = function(data){
        for(var e = 0; e < serverErrorMessageListeners.length; e++){
            try{
                serverErrorMessageListeners[e](data);
            }
            catch(err){
                LODE_LOG("Callback for server errors failed for " + err.message);
            }
        }
    };

};




var Queue = function(){

    var queueItems = [];
    var currentQueueItem = "";
    var currentQueueItemUpdateListeners = [];
	var size = 0;

    this.setCurrentQueueItem = function(num){
        currentQueueItem = num;
        this.fireCurrentQueueItemEvents();
    };

    this.getCurrentQueueItem = function(){
        return currentQueueItem;
    };

    this.size = function(){
        return queueItems.length;;
    };

    this.setQueueItems = function(aqueueItems){
        queueItems = aqueueItems;
    };

    this.getQueueItems = function(){
        return queueItems;
    };

    this.addQueueItem = function(aqueueItem){
        queueItems.push(aqueueItem);
    };


    this.registerCurrentQueueItemUpdateListener = function(listener){
        currentQueueItemUpdateListeners.push(listener);
    };

    this.deregisterCurrentQueueItemUpdateListener = function(listener){
        currentQueueItemUpdateListeners.splice(currentQueueItemUpdateListeners.indexOf(listener), 1);
    };

    this.fireCurrentQueueItemEvents = function(){
        for(var i = 0; i < currentQueueItemUpdateListeners.length; i++){
            try{
                currentQueueItemUpdateListeners[i]();
            }
            catch(err){
                LODE_LOG("Callback for current queue item failed for " +  err.message);
            }
        }
    };

};



var QueueItem = function(res, title, artist, albumArt, attributes){

    var res = res;
    var title = title;
    var artist = artist;
    var albumArt = albumArt;
    var attributes = attributes;

    this.getRes = function(){
        return res;
    };

    this.getTitle = function(){
        return title;
    };

    this.getArtist = function(){
        return artist;
    };

    this.getAlbumArt = function(){
        return albumArt;
    };

    this.getSttributes = function(){
        return attributes;
    }

};



var Track = function(){
	
	var id = "";
    var name = "";
    var album = "";
    var artist = "";
    var albumArt = "";
    var progress = "";
    var duration = "";
    var remaining = "";
    var trackProgressListeners = [];
    var zone = "";
	var stream = false;

   this.update =  function(azone, aid, aname, aalbum, aartist, aalbumArt, duration) {
        zone = azone;
        id = aid;
        name = aname;
        album = aalbum;
        artist = aartist;
        albumArt = aalbumArt;
        progress = 0;
		LODE_LOG("Duration be sent for track " + duration);
		if(duration > 0)
		{
			stream = false;
		}
		else
		{
			stream = true;
		}
        remaining = "";
    };
	
	
    this.setZone = function(aZone) {
        zone = aZone;
    };

    this.getZone = function(){
        return zone;
    };

    this.setId = function (aid) {
        id = aid;
    };

    this.getId = function(){
        return id;
    };

    this.setName = function(aname){
		name = aname;
	};

    this.getName = function(){
        return name;
    };

    this.setAlbum = function(aalbum){
		album = aalbum;
	};

    this.getAlbum = function(){
        return album;
    };

    this.setArtist = function(aartist){
		artist = aartist;
	};

    this.getArtist = function(){
        return artist;
    };

    this.setAlbumArt = function(aalbumArt){
		albumArt = aalbumArt;
	};

    this.getAlbumArt = function(){
        return albumArt;
    };

	this.isAStream = function(){
		return stream;
	};
	
    this.setProgress = function(aprogress) {
        progress = aprogress;
    };

    this.getProgress = function(){
        return progress;
    };

    this.setDuration = function(aduration){
        duration =  aduration;
    };

    this.getDuration = function(){
        return duration;
    };

    this.setRemaining = function(aremaining){
        remaining = aremaining;
    };

    this.getRemaining = function(){
        return remaining;
    };

    this.updateProgress = function(avalue, astart, aend){
        this.setProgress(avalue);
        this.setDuration(astart);
        this.setRemaining(aend);

        //Fire the update events.
        this.fireTrackProgressEvents();

    };

    this.registerTrackProgressListener = function(listener){
        trackProgressListeners.push(listener);
    };

    this.deregisterTrackProgressListener = function(listener){
        trackProgressListeners.splice(trackProgressListeners.indexOf(listener), 1);
    };

    this.fireTrackProgressEvents = function(){
        for(var e = 0; e <  trackProgressListeners.length; e++){
            try{
                trackProgressListeners[e](zone);
            }
            catch(e){
                LODE_LOG("Callback for track progress failed for " +  e.message);
            }
        }
    };

};



var Zone = function(zoneName, lodeAPI, model){

    var name = zoneName;
    var api = lodeAPI;
    var volume = 0;
    var volumeUpdateListeners = [];
    var mute = 0;
    var muteUpdateListeners =[ ];
    var transportState = "";
    var transportStateUpdateListeners = [];
    var playMode = "";
    var playModeUpdateListeners = [];
    var children = [];
    var queue = new Queue();
    var queueUpdateListeners = [];
    var track = new Track();
    var trackUpdateListeners = [];
	var nextTrackUpdateListeners = [];
    var nextTrack = "";
    var isMaster = true;
	var parent = null;
    var zoneUpdateListeners = [];
    var zoneRemovalListeners = [];
	var Model = model;

    //Used for debug
    this.id = name;
    var volumeTimer = "";
    var volumeTimerActive = false;
    var tempVolume = 0;

    this.update = function(group){

        try {
			
			parent = null;
            isMaster = true;
            var updated = false;
			
            for (var i = 0; i < children.length; i++) {
                if (group.indexOf(children[i].getName()) == -1) {
                    updated = true;
                    children.splice(i, 1);
                }
            }

            if (group.length > 1) {
                for (var i = 1; i < group.length; i++) {

                    var zone = Model.getZone(group[i]);

                    if (!zone) {
                        zone = new Zone(group[i],api);
                        Model.addZone(zone);
                        Model.queryChildVolume(zone.getName());
                    }

                    if (!this.hasChild(zone.getName())) {
                        this.addChild(zone);
                        updated = true;
                    }

                }
            }
            else {
                children = [];
            }

            if (updated) {
                this.fireZoneUpdateEvents();
            }
        }
        catch(err){
            LODE_LOG("Error parsing zones " +  err.message)
        }
    };

    this.setName = function(aname){
        name = aname;
    };

    this.getName = function(){
        return name;
    };

    this.setVolume = function(avol){
        volume = avol;
        this.fireVolumeUpdateEvents();
    };

    this.getVolume = function(){
        return volume;
    };

    this.setMute = function(amute){
        mute = amute;
        this.fireMuteUpdateEvents();
    };

    this.getMute = function(){
        return mute;
    };

    this.isFixedVolume = function(){
        return volume == -1;
    };

    this.setTransportState = function(astate){
        transportState = astate;
        this.fireTransportStateUpdateEvents();
    };

    this.getTransportState = function(){
      return transportState;
    };

    this.setPlayMode = function(amode){
        playMode = amode;
        this.firePlayModeUpdateEvents();
    };

    this.getPlayMode = function(){
      return playMode;
    };

    this.hasChild = function(zone){

        for(var i = 0; i < children.length; i++){
            if(children[i].getName() == zone){
                return true;
            }
        }
        return false;
    };

    this.getChildren = function(){
        return children;
    };

    this.getChildrenAndParent = function(){
        var zonesArr = [];
        zonesArr.push(this);

        for(var zone in children){
            zonesArr.push(children[zone]);
        }

        return zonesArr;
    };

	this.setParent = function(aparent){
		parent = aparent;
	};

	this.getParent = function(){
		return parent;
	};

    this.addChild = function(child){
        if(!this.hasChild(child.name)) {
            child.setMaster(false);
			child.setParent(this);
            children.push(child);
        }
    };

    this.setQueue = function(aqueue){
        queue = aqueue;
        this.fireQueueUpdateEvents();
    };

    this.getQueue = function(){
        return queue;
    };

    this.updateTrack = function(zone, id, name, album, artist, albumArt, duration){
        track.update(zone, id, name, album, artist, albumArt,  duration);
        this.fireTrackUpdateEvents();
    };

    this.setTrack = function(atrack){
        track = atrack;
        this.fireTrackUpdateEvents();
    };

    this.getTrack = function(){
        return track;
    };

    this.setNextTrack = function(atrack) {
        nextTrack = atrack;
		LODE_LOG("Setting next track to: " + atrack);
		this.fireNextTrackUpdateEvents();
    };

    this.getNextTrack = function(){
        return nextTrack;
    };

    this.setMaster = function(aisMaster){
        isMaster = aisMaster;
    };

    this.isMaster = function(){
        return isMaster;
    };

    this.isCurrentZone = function(){
        return name == Model.getCurrentZone().getName();
    };

    ///////////   API Commands   /////////

    this.sendPlay = function(){
        api.play(name)
    };

    this.sendPause = function(){
        api.pause(name);
    };

    this.sendPreviousTrack = function(){
        api.previousTrack(name);
    };

    this.sendNextTrack = function(){
        api.nextTrack(name);
    };

    this.sendSeek = function(percentage){
        api.seek(name, percentage);
    };

    this.sendPlayNow = function(res){
        api.playNow(name, res);
    };

    this.sendPlayNext = function(res){
        api.playNext(name, res);
    };

    this.sendReplaceQueue = function(res){
        api.replaceQueue(name, res);
    };

    this.sendAddToQueue = function(res){
        api.addToQueue(name, res);
    };

    this.sendAddToFav = function(res){
        api.addToFav(name, res);
    };

    this.sendDeleteFav = function(res){
        api.deleteFav(name, res);
    };

    this.sendDeletePlaylist = function(res){
        api.deletePlaylist(name, res);
    };

    this.sendMute = function(state){
        api.updateMuteForZone(name, state);
    };


    var volumeUpdater = function(){
        api.setVolume(name, tempVolume)
        volumeTimerActive = false;
    };

    this.sendVolume = function(newVolume, throttled){
		if(throttled){
			LODE_LOG('----------  in zone set volume for zone:' + newVolume);
			tempVolume = newVolume;
			if(!volumeTimerActive) {
				volumeTimerActive = true;
				volumeTimer = setTimeout(volumeUpdater, 250);
			}
		}
		else{
			api.setVolume(name, newVolume);
		}
    };

    this.sendGroupedVolume = function(newVolume){
        api.setGroupedVolume(name, newVolume);
    };

    this.sendQueryVolume = function(){
        api.queryVolumeForZone(name);
    };

    this.sendPartyMode = function(){
        api.setPartyModeForZone(name);
    };
	
	this.queryCurrentQueueItem = function(){
		 api.queryCurrentQueueItemForZone(name);
	}

    this.removeAllMembers = function() {
        api.removeAllMembers(name);
    };

    this.sendPlayMode = function(mode){
        api.updatePlayModeForZone(name, mode);
    };

    this.sendRemoveMember = function(member){
        api.removeZoneMember(member);
    };

    this.sendAddMember = function(member){
        api.addZoneMember(name,member);
    };

    this.sendClearQueue = function(){
        api.clearQueue(name);
    };

    this.sendSaveQueue = function(playlistName){
        api.saveQueue(name, playlistName);
    };

    this.sendReorderTrackInQueue = function(item, newPos){
        api.reorderTrackInQueue(name, item, newPos);
    };

    this.sendPlayQueueItem = function(item){
        api.playQueue(name, item);
    };

    this.sendRemoveFromQueue = function(item){
        api.removeFromQueue(name, item);
    };

    this.sendSearch = function(path, criteria, term, index, total){
        api.doSearch(name, path, criteria, term, index, total);
    };


    //////////   Events   //////////
    this.registerTransportStateUpdateListener = function(listener){
        transportStateUpdateListeners.push(listener);
    };

    this.deregisterTransportStateUpdateListener = function(listener){
        transportStateUpdateListeners.splice(transportStateUpdateListeners.indexOf(listener), 1);
    };

    this.fireTransportStateUpdateEvents = function(){
        for(var ev in transportStateUpdateListeners){
            try{
                transportStateUpdateListeners[ev](name);
            }
            catch(err){
                LODE_LOG("Callback for transport failed for " +  err.message);
            }
        }
    };

    this.registerVolumeUpdateListener = function(listener){
        volumeUpdateListeners.push(listener);
        listener(name);
    };

    this.deregisterVolumeUpdateListener = function(listener){
        volumeUpdateListeners.splice(volumeUpdateListeners.indexOf(listener), 1);
    };

    this.fireVolumeUpdateEvents = function(){
        for(var e = 0; e < volumeUpdateListeners.length; e++){
            try{
               volumeUpdateListeners[e](name);
            }
            catch(err){
                LODE_LOG("Callback for volume failed for " + err.message);
            }
        }
    };

    this.registerMuteUpdateListener = function(listener){
        muteUpdateListeners.push(listener);
    };

    this.deregisterMuteUpdateListener = function(listener){
        muteUpdateListeners.splice(muteUpdateListeners.indexOf(listener), 1);
    };

    this.fireMuteUpdateEvents = function(){
        for(var e = 0; e < muteUpdateListeners.length; e++){
            try{
                muteUpdateListeners[e](name);
            }
            catch(err){
                LODE_LOG("Callback for mute failed for " +  err.message);
            }
        }
    };

    this.registerPlayModeUpdateListener = function(listener){
        playModeUpdateListeners.push(listener);
    };

    this.deregisterPlayModeUpdateListener = function(listener){
        playModeUpdateListeners.splice(playModeUpdateListeners.indexOf(listener), 1);
    };

    this.firePlayModeUpdateEvents = function(){
        for(var e = 0; e < playModeUpdateListeners.length; e++){
            try{
                playModeUpdateListeners[e](name);
            }
            catch(err){
                LODE_LOG("Callback for play mode failed for " +  err.message);
            }
        }
    };

    this.registerQueueUpdateListener = function(listener){
        queueUpdateListeners.push(listener);
    };

    this.deregisterQueueUpdateListener = function(listener){
        queueUpdateListeners.splice(queueUpdateListeners.indexOf(listener), 1);
    };

    this.fireQueueUpdateEvents = function(){
        for(var e = 0; e < queueUpdateListeners.length; e++){
            try{
                queueUpdateListeners[e]();
            }
            catch(err){
                LODE_LOG("Callback for queue failed for " +  err.message);
            }
        }
    };
	
	 this.registerNextTrackUpdateListener = function(listener){
        nextTrackUpdateListeners.push(listener);
    };

    this.deregisterNextTrackUpdateListener = function(listener){
        nextTrackUpdateListeners.splice(nextTrackUpdateListeners.indexOf(listener), 1);
    };

    this.fireNextTrackUpdateEvents = function(){
        for(var e = 0; e < nextTrackUpdateListeners.length; e++){
            try{
                nextTrackUpdateListeners[e]();
            }
            catch(err){
                LODE_LOG("Callback for next track in Zone  for " +  err.message);
            }
        }
    };

    this.registerTrackUpdateListener = function(listener){
        trackUpdateListeners.push(listener);
    };

    this.deregisterTrackUpdateListener = function(listener){
        trackUpdateListeners.splice(trackUpdateListeners.indexOf(listener), 1);
    };

    this.fireTrackUpdateEvents = function(){
        for(var e = 0; e < trackUpdateListeners.length; e++){
            try{
                trackUpdateListeners[e]();
            }
            catch(err){
                LODE_LOG("Callback for track update failed for " +  err.message);
            }
        }
    };


    this.registerZoneUpdateListener = function(listener){
        zoneUpdateListeners.push(listener);
    };

    this.deregisterZoneUpdateListener = function(listener){
        zoneUpdateListeners.splice(zoneUpdateListeners.indexOf(listener), 1);
    };

    this.fireZoneUpdateEvents = function(){
        for(var e = 0; e < zoneUpdateListeners.length; e++){
            try{
                zoneUpdateListeners[e]();
            }
            catch(err){
                LODE_LOG("Callback for track update failed for " +  err.message);
            }
        }
    };


    this.registerZoneRemovalListener = function(listener){
        zoneRemovalListeners.push(listener);
    };

    this.deregisterZoneRemovalListener = function(listener){
        zoneRemovalListeners.splice(zoneRemovalListeners.indexOf(listener), 1);
    };

    this.fireZoneRemovalEvents = function(){
        for(var e = 0; e < zoneRemovalListeners.length; e++){
            try{
                zoneRemovalListeners[e]();
            }
            catch(err){
                LODE_LOG("Callback for track update failed for " + err.message);
            }
        }
    };

};


var Parser = function() {


    this.parseResponse = function(str) {

        var obj = {text: str};
        var params = [];

        while (obj.text.length != 0) {

            var value = this.getNextParameter(obj);

            if (value.length > 0) {
                params.push(value);
            }
            else {
                params.push('');
            }

        }
        return params;
    };


    this.findCloseBracePosition = function(theString) {
        var braceCount = 0;
        var doubleQuotes = false;

        for (var i = 0; i < theString.length; i++) {

            if (theString.charAt(i) == '"' && theString.charAt(i - 1) == '"') {
                if (theString.charAt(i - 2) == ',' || theString.charAt(i - 2) == '{') {
                    doubleQuotes = true;
                }
                if (theString.charAt(i + 1) == ',' || theString.charAt(i + 1) == '}') {
                    doubleQuotes = false;
                }
            }

            if (!doubleQuotes) {
                if (theString.charAt(i) == '{') {
                    braceCount++;
                }
                if (theString.charAt(i) == '}') {
                    braceCount--;
                }
            }
            if (theString.charAt(i) == '}' && !doubleQuotes && braceCount == 0) {
                return i;
            }
        }
        return false;
    };

    this.getArrayElements = function(obj) {

        var params = [];
        var newObj = {text: ''};

        var end = this.findCloseBracePosition(obj.text);
        if (!end) {
            return '';
        }

        newObj.text = obj.text.substring(1, end);
        obj.text = obj.text.substring(end + 1, obj.text.length);

        //remove preceeding commas for array items.
        if (obj.text.charAt(0) == ',') {
            obj.text = obj.text.substring(1, obj.text.length);
        }

        while (newObj.text.length != 0) {
            if (newObj.text.charAt(0) == '{') {
                var value = this.getArrayElements(newObj);
                params.push(value);
            }
            else {
                var value = this.getNextParameter(newObj);
                if (value.length > 0) {
                    params.push(value);
                }
                else {
                    params.push('');
                }
            }
        }
        return params;
    };


    this.nthIndex = function(str, pat, n) {
        var L = str.length, i = -1;
        while (n-- && i++ < L) {
            i = str.indexOf(pat, i);
        }
        return i;
    };


    this.getNextParameter = function(obj) {

        var param = '';

        for (var i = 0; i < obj.text.length; i++) {

            var startsWithDoubleQuote = obj.text.startsWith('""');
            var secondDoubleQuotesPosition = this.nthIndex(obj.text, '""', 2);

            if (startsWithDoubleQuote && (startsWithDoubleQuote || secondDoubleQuotesPosition != 1 )) {
                var end = obj.text.startsWith('"""') ? this.nthIndex(obj.text, '""', 3) : secondDoubleQuotesPosition;
                param = obj.text.substring(2, end);
                obj.text = obj.text.substring(param.length + 5, obj.text.length);
                return param;
            }

            if (startsWithDoubleQuote && secondDoubleQuotesPosition == 1) {
                obj.text = obj.text.substring(param.length + 5, obj.text.length);
                return '';
            }


            if (obj.text.startsWith('{')) {
                return this.getArrayElements(obj);
            }

            if (obj.text.charAt(i) == ',' || obj.text.charAt(i) == '}') {
                obj.text = obj.text.substring(param.length + 1, obj.text.length);
                return param;
            }
            else if (i == (obj.text.length - 1)) {
                param = obj.text;
                obj.text = obj.text.substring(param.length + 1, obj.text.length);
                return param;
            }

            param += obj.text.charAt(i);

        }
        return param;
    }

};



var API = function(module) {

    this.services = {};
    this.browseRequestAmount = 0;
    this.parser = new Parser();
    this.debugSocket = true;
	this.module = module;
	
	var obj = this;
	
	
	this.send = function(string){
		LODE_LOG('Socket Send:' + string);
		CF.send(this.module.sysName, string + "\x0D\x0A");
	};

    this.sendInitialisationQueries = function(){
		LODE_LOG('Sending API init commands');
        this.send('?VERSION');
        this.send('?ZONES');
        this.send('?SEARCHCRITERIA');
    };

    this.init = function (browseRequestAmount) {
		this.browseRequestAmount = browseRequestAmount;	
		LODE_LOG('In API init');
	};		

    this.getServiceIdForBrowseId = function (id) {
        for (var service in this.services) {
            if (id.indexOf(service) != -1) {
                return service;
            }
        }
        return false;
    };

    this.setServices = function (data) {
        this.services = {};
        for (var i = 0; i < data.size(); i++) {
            this.services[data.getBrowseItemForIndex(i).getRes()] =  data.getBrowseItemForIndex(i);
        }
    };

    this.setVolume = function (zone, volume) {
        this.send('#VOLUME,' + zone + ',' + volume);
    };

    this.setGroupedVolume = function(zone, volume){
        this.send('#GROUPVOLUME,'  + zone + ',' + volume);
    };

    this.queryPlayModeForZone = function (zone) {
        this.send('?PLAYMODE,' + zone);
    };

    this.queryQueueForZone = function (zone) {
        this.send('?QUEUE,' + zone + ',0,' + 1000);
    };

    this.queryTransportStateForZone = function (zone) {
        this.send('?TRANSPORT,' + zone);
    };

    this.queryTrackForZone = function (zone) {
        this.send('?TRACK,' + zone);
    };

	this.queryNextTrackForZone = function (zone) {
        this.send('?NEXTTRACK,' + zone);
    };
	
    this.queryBrowse = function (zone, id, index) {
        this.send('#BROWSE,' + zone + ',' + id + ',' + index + ',' + this.browseRequestAmount);
    };

    this.seek = function (zone, percentage) {
        this.send('#SEEK,' + zone + ',' + percentage + ',' + 100);
    };

    this.play = function (zone) {
        this.send("#PLAY," + zone);
    };

    this.pause = function (zone) {
        this.send("#PAUSE," + zone);
    };

    this.playNow = function (zone, res) {
        this.send("#PLAYNOW," + zone + "," + res)
    };

    this.playNext = function (zone, res) {
        this.send("#PLAYNEXT," + zone + "," + res)
    };

    this.replaceQueue = function (zone, res) {
        this.send("#REPLACEQUEUE," + zone + "," + res)
    };

    this.addToQueue = function (zone, res) {
        this.send("#ADDTOQUEUE," + zone + "," + res)
    };

    this.addToFav = function (zone, res) {
        this.send("#ADDTOFAV," + zone + "," + res)
    };

    this.deleteFav = function (zone, res) {
        this.send("#DELETEFAV," + zone + "," + res)
    };

    this.deletePlaylist = function (zone, res) {
        this.send("#DELETEPLAYLIST," + zone + "," + res)
    };

    this.updateMuteForZone = function (zone, state) {
        this.send("#MUTE," + zone + "," + state);
    };

    this.updatePlayModeForZone = function (zone, mode) {
        this.send("#PLAYMODE," + zone + "," + mode);
    };

    this.setPartyModeForZone = function (zone) {
        this.send("#PARTYMODE," + zone);
    };

    this.clearQueue = function (zone) {
        this.send("#CLEARQUEUE," + zone);
    };

    this.saveQueue = function (zone, playlist) {
        this.send("#SAVEQUEUE," + zone + "," + playlist);
    };

    this.reorderTrackInQueue = function (zone, item, pos) {
        this.send("#REORDERTRACKINQUEUE," + zone + "," + item + "," + pos);
    };

    this.removeFromQueue = function (zone, item) {
        this.send("#REMOVEFROMQUEUE," + zone + "," + item);
    };

    this.doSearch = function (zone, path, criteria, term, index, total) {
        this.send("#SEARCH," + zone + "," + path + "," + criteria + ",\"\"" + term + "\"\"" + "," + index + "," + total);
    };

    this.playQueue = function (zone, item) {
        this.send("#PLAYQUEUE," + zone + "," + item);
    };

    this.queryVolumeForZone = function (zone) {
        this.send('?VOLUME,' + zone);
    };

    this.queryMuteForZone = function (zone) {
        this.send('?MUTE,' + zone);
    };

    this.queryCurrentQueueItemForZone = function (zone) {
        this.send('?CURRENTQUEUEITEM,' + zone);
    };

    this.nextTrack = function (zone) {
        this.send('#NEXT,' + zone);
    };

    this.previousTrack = function (zone) {
        this.send('#PREVIOUS,' + zone);
    };

    this.removeZoneMember = function (zone) {
        this.send("#REMOVEMEMBER," + zone);
    };

    this.removeAllMembers = function (zone) {
        this.send("#REMOVEALLMEMBERS," + zone);
    };

    this.addZoneMember = function (zone, child) {
        this.send("#ADDMEMBER," + zone + "," + child);
    };


    //////// DATA Received Events  ////////
    var zonesEventListeners = [];
    var versionEventListeners = [];
    var playersEventListeners = [];
    var transportStateEventListeners = [];
    var trackEventListeners = [];
	var  nextTrackEventListeners = [];
    var trackProgressEventListeners = [];
    var volumeEventListeners = [];
    var muteEventListeners = [];
    var browseEventListeners = [];
    var queueEventListeners = [];
    var currentQueueItemEventListeners = [];
    var queueChangedEventListeners = [];
    var searchCriteriaEventListeners = [];
    var playModeEventListeners = [];
    var messageEventListeners = [];


    this.registerZoneEventListener = function(listener){
        zonesEventListeners.push(listener);
    };

    this.deregisterZoneEventListener = function(listener){
        zonesEventListeners.splice(zonesEventListeners.indexOf(listener), 1);
    };

    this.fireZoneEvents = function(responseData){
        for(var e = 0; e < zonesEventListeners.length; e++){
            try{
                zonesEventListeners[e](responseData);
            }
            catch(err){
                LODE_LOG("Callback for zone data received failed for " + err.message);
            }
        }
    };


    this.registerVersionEventListener = function(listener){
        versionEventListeners.push(listener);
    };

    this.deregisterVersionEventListener = function(listener){
        versionEventListeners.splice(versionEventListeners.indexOf(listener), 1);
    };

    this.fireVersionEvents = function(responseData){
        for(var e = 0; e < versionEventListeners.length; e++){
            try{
                versionEventListeners[e](responseData);
            }
            catch(err){
                LODE_LOG("Callback for version data received failed for " + err.message);
            }
        }
    };


    this.registerPlayersEventListener = function(listener){
        playersEventListeners.push(listener);
    };

    this.deregisterPlayersEventListener = function(listener){
        playersEventListeners.splice(playersEventListeners.indexOf(listener), 1);
    };

    this.firePlayersEvents = function(responseData){
        for(var e = 0; e < playersEventListeners.length; e++){
            try{
                playersEventListeners[e](responseData);
            }
            catch(err){
                LODE_LOG("Callback for players data received failed for " + err.message);
            }
        }
    };

    this.registerTransportStateEventListener = function(listener){
        transportStateEventListeners.push(listener);
    };

    this.deregisterTransportStateEventListener = function(listener){
        transportStateEventListeners.splice(transportStateEventListeners.indexOf(listener), 1);
    };

    this.fireTransportStateEvents = function(responseData){
        for(var e = 0; e < transportStateEventListeners.length; e++){
            try{
                transportStateEventListeners[e](responseData);
            }
            catch(err){
                LODE_LOG("Callback for transport state received failed for " + err.message);
            }
        }
    };
	
	
	
    this.registerNextTrackEventListener = function(listener){
        nextTrackEventListeners.push(listener);
    };

    this.deregisterNextTrackEventListener = function(listener){
        nextTrackEventListeners.splice(nextTrackEventListeners.indexOf(listener), 1);
    };

    this.fireNextTrackEvents = function(responseData){
        for(var e = 0; e < nextTrackEventListeners.length; e++){
            try{
                nextTrackEventListeners[e](responseData);
            }
            catch(err){
                LODE_LOG("Callback for next track data received failed for " + err.message);
            }
        }
    };


    this.registerTrackEventListener = function(listener){
        trackEventListeners.push(listener);
    };

    this.deregisterTrackEventListener = function(listener){
        trackEventListeners.splice(trackEventListeners.indexOf(listener), 1);
    };

    this.fireTrackEvents = function(responseData){
        for(var e = 0; e < trackEventListeners.length; e++){
            try{
                trackEventListeners[e](responseData);
            }
            catch(err){
                LODE_LOG("Callback for track data received failed for " + err.message);
            }
        }
    };


    this.registerTrackProgressEventListener = function(listener){
        trackProgressEventListeners.push(listener);
    };

    this.deregisterTrackProgressEventListener = function(listener){
        trackProgressEventListeners.splice(trackProgressEventListeners.indexOf(listener), 1);
    };

    this.fireTrackProgressEvents = function(responseData){
        for(var e = 0; e < trackProgressEventListeners.length; e++){
            try{
                trackProgressEventListeners[e](responseData);
            }
            catch(err){
                LODE_LOG("Callback for trackprogress data received failed for " + err.message);
            }
        }
    };


    this.registerVolumeEventListener = function(listener){
        volumeEventListeners.push(listener);
    };

    this.deregisterVolumeEventListener = function(listener){
        volumeEventListeners.splice(volumeEventListeners.indexOf(listener), 1);
    };

    this.fireVolumeEvents = function(responseData){
        for(var e = 0; e < volumeEventListeners.length; e++){
            try{
                volumeEventListeners[e](responseData);
            }
            catch(err){
                LODE_LOG("Callback for volume data received failed for " + err.message);
            }
        }
    };


    this.registerMuteEventListener = function(listener){
        muteEventListeners.push(listener);
    };

    this.deregisterMuteEventListener = function(listener){
        muteEventListeners.splice(muteEventListeners.indexOf(listener), 1);
    };

    this.fireMuteEvents = function(responseData){
        for(var e = 0; e < muteEventListeners.length; e++){
            try{
                muteEventListeners[e](responseData);
            }
            catch(err){
                LODE_LOG("Callback for mute data received failed for " + err.message);
            }
        }
    };


    this.registerBrowseEventListener = function(listener){
        browseEventListeners.push(listener);
    };

    this.deregisterBrowseEventListener = function(listener){
        browseEventListeners.splice(browseEventListeners.indexOf(listener), 1);
    };

    this.fireBrowseEvents = function(responseData){
        for(var e = 0; e < browseEventListeners.length; e++){
            try{
                browseEventListeners[e](responseData);
            }
            catch(err){
                LODE_LOG("Callback for browse data received failed for " + err.message);
            }
        }
    };


    this.registerQueueEventListener = function(listener){
        queueEventListeners.push(listener);
    };

    this.deregisterQueueEventListener = function(listener){
        queueEventListeners.splice(queueEventListeners.indexOf(listener), 1);
    };

    this.fireQueueEvents = function(responseData){
        for(var e = 0; e < queueEventListeners.length; e++){
            try{
                queueEventListeners[e](responseData);
            }
            catch(err){
                LODE_LOG("Callback for queue data received failed for " + err.message);
            }
        }
    };


    this.registerCurrentQueueItemEventListener = function(listener){
        currentQueueItemEventListeners.push(listener);
    };

    this.deregisterCurrentQueueItemEventListener = function(listener){
        currentQueueItemEventListeners.splice(currentQueueItemEventListeners.indexOf(listener), 1);
    };

    this.fireCurrentQueueItemEvents = function(responseData){
        for(var e = 0; e < currentQueueItemEventListeners.length; e++){
            try{
                currentQueueItemEventListeners[e](responseData);
            }
            catch(err){
                LODE_LOG("Callback for currentQueueItem data received failed for " + err.message);
            }
        }
    };


    this.registerQueueChangedEventListener = function(listener){
        queueChangedEventListeners.push(listener);
    };

    this.deregisterQueueChangedEventListener = function(listener){
        queueChangedEventListeners.splice(queueChangedEventListeners.indexOf(listener), 1);
    };

    this.fireQueueChangedItemEvents = function(responseData){
        for(var e = 0; e < queueChangedEventListeners.length; e++){
            try{
                queueChangedEventListeners[e](responseData);
            }
            catch(err){
                LODE_LOG("Callback for queue changed data received failed for " + err.message);
            }
        }
    };


    this.registerSearchCritereaEventListener = function(listener){
        searchCriteriaEventListeners.push(listener);
    };

    this.deregisterSearchCritereaEventListener = function(listener){
        searchCriteriaEventListeners.splice(searchCriteriaEventListeners.indexOf(listener), 1);
    };

    this.fireSearchCritereaEvents = function(responseData){
        for(var e = 0; e < searchCriteriaEventListeners.length; e++){
            try{
                searchCriteriaEventListeners[e](responseData);
            }
            catch(err){
                LODE_LOG("Callback for search criterea data received failed for " + err.message);
            }
        }
    };


    this.registerPlayModeEventListener = function(listener){
        playModeEventListeners.push(listener);
    };

    this.deregisterPlayModeEventListener = function(listener){
        playModeEventListeners.splice(playModeEventListeners.indexOf(listener), 1);
    };

    this.firePlayModeEvents = function(responseData){
        for(var e = 0; e < playModeEventListeners.length; e++){
            try{
                playModeEventListeners[e](responseData);
            }
            catch(err){
                LODE_LOG("Callback for play mode data received failed for " + err.message);
            }
        }
    };


    this.registerMessageEventListener = function(listener){
        messageEventListeners.push(listener);
    };

    this.deregisterMessageEventListener = function(listener){
        messageEventListeners.splice(messageEventListeners.indexOf(listener), 1);
    };

    this.fireMessageEvents = function(responseData){
        for(var e = 0; e < messageEventListeners.length; e++){
            try{
                messageEventListeners[e](responseData);
            }
            catch(err){
                LODE_LOG("Callback for message data received failed for " + err.message);
            }
        }
    };


    this.handleData = function (data) {

        var stringData = data.trim();
        var index = stringData.indexOf(',');

        var responseType = stringData.substr(1, index - 1);
        var responseData = this.parser.parseResponse(stringData.substr(index + 1));

        if (this.debugSocket){
            if ("TRACKPROGRESS" != responseType && "" != responseType) {
                LODE_LOG('RECEIVED COMMAND ' + responseType + " " + stringData);
            }
        }

        try {

            switch (responseType) {
                case 'ZONES':
                    this.fireZoneEvents(responseData);
                    break;
                case 'VERSION':
                    this.fireVersionEvents(responseData);
                    break;
                case 'PLAYERS':
                    this.firePlayersEvents(responseData);
                    break;
                case 'TRANSPORT':
                    this.fireTransportStateEvents(responseData);
                    break;
                case 'TRACK':
                    this.fireTrackEvents(responseData);
                    break;
				case 'NEXTTRACK':
                    this.fireNextTrackEvents(responseData);
                    break;
                case 'TRACKPROGRESS':
                    this.fireTrackProgressEvents(responseData);
                    break;
                case 'VOLUME':
                    this.fireVolumeEvents(responseData);
                    break;
                case 'MUTE':
                    this.fireMuteEvents(responseData);
                    break;
                case 'BROWSE':
					LODE_LOG("--------------Browse firing from API-----------------")
                    this.fireBrowseEvents(responseData);
                    break;
                case 'QUEUE':
                    this.fireQueueEvents(responseData);
                    break;
                case 'CURRENTQUEUEITEM':
                    this.fireCurrentQueueItemEvents(responseData);
                    break;
                case 'QUEUECHANGED':
                    this.fireQueueChangedItemEvents(responseData);
                    break;
                case 'SEARCHCRITERIA':
                    this.fireSearchCritereaEvents(responseData);
                    break;
                case 'PLAYMODE':
                    this.firePlayModeEvents(responseData);
                    break;
                case 'MESSAGE':
                    this.fireMessageEvents(responseData);
                    break;
                default:
                    break;
            }
        }
        catch(e){
            LODE_LOG("Error processing API response data for " + responseType + " with message: " + e.message);
        }

    };

};



